from __future__ import annotations
import cv2
import numpy as np
import copy
from PIL import Image, ImageDraw

from config import *
import checker
import parsers
import utils

def resize_image_respect_aspect_ratio(image, width, height, interpolation=cv2.INTER_CUBIC):
    if (width is not None) and (height is None):
        ratio = width / image.shape[1]
        height = int(image.shape[0] * ratio)
        image = cv2.resize(image, (width, height), interpolation=interpolation)
    elif (height is not None) and (width is None):
        ratio = height / image.shape[0]
        width = int(image.shape[1] * ratio)
        image = cv2.resize(image, (width, height), interpolation=interpolation)
    elif (height is not None) and (width is not None):
        image, *_ = resize_image_respect_aspect_ratio(image, width=width, height=None)
        image, *_ = resize_image_respect_aspect_ratio(image, width=None, height=height)

    h, w, _ = image.shape
    return image, h, w


def resize_universal_output_image(output_image:np.ndarray, max_output_image_size_wh:tuple):
    checker.assert_types([output_image, max_output_image_size_wh],
                         [np.ndarray, tuple], ["output_image", "max_output_image_size_wh"], [0, 1])
    if max_output_image_size_wh is None:
        return output_image

    # Resize output image if user wants it
    max_w_out, max_h_out = max_output_image_size_wh
    final_h, final_w, _ = output_image.shape

    if max_w_out and (final_w < max_w_out):
        max_w_out = None
    if max_h_out and (final_h < max_h_out):
        max_h_out = None

    if max_w_out and max_h_out:
        output_image, *_ = resize_image_respect_aspect_ratio(output_image, width=max_w_out, height=max_h_out)
    elif max_w_out or max_h_out:
        output_image, *_ = resize_image_respect_aspect_ratio(output_image, width=max_w_out, height=max_h_out)

    return output_image


def center_pad_images_into_nparray(source):
    max_height = max(img.shape[0] for img in source)
    max_width = max(img.shape[1] for img in source)

    padded_images = []
    for img in source:
        height, width, _ = img.shape
        pad_height = (max_height - height) // 2
        pad_width = (max_width - width) // 2

        pad_top = pad_height
        pad_bottom = max_height - (height + pad_height)
        pad_left = pad_width
        pad_right = max_width - (width + pad_width)

        padded_img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right), (0, 0)), mode='constant')
        padded_images.append(padded_img)

    return np.array(padded_images)


def draw_text_cv2(
        image_source: ImageType,
        text: str,
        position: Union[str, tuple] = "upper_left_corner",
        font: int = cv2.FONT_HERSHEY_DUPLEX,
        font_scale: int = None,
        color: tuple = None,
        thickness=None,
        background_color: Union[str, tuple] = "auto",
        line_type: int = cv2.LINE_AA,
        padding_pixels: int = 10,
        deep_copy:bool=True
):
    """
    A function designed to draw `text` on `image` with minimum effort.

    # EXAMPLE
    >> image = np.zeros((512, 1024, 3)).astype(np.uint8)
    >> draw_text(image, "Hello 123", "bottom_left_corner", padding_pixels=100)

    :param image_source: An image. Can handle most image sources: ndarray, url, path, torch, PIL. For a full description see: TODO insert link
    :param text: The text string to be drawn on the image.
    :param position: The position of the text on the image. It can be specified as a string:
                     ("upper_left_corner", "upper_right_corner", "bottom_left_corner", "bottom_right_corner", or "center")
                     or as a tuple of (x, y) coordinates.
    :param font: The font type to be used for the text. It should be one of the predefined constants from the cv2 library (e.g., cv2.FONT_HERSHEY_DUPLEX).
    :param font_scale: The scale factor for the font size. If not provided, it will be calculated based on the image size.
    :param color: The color of the text, specified as a tuple of RGB values. If not provided, it will be automatically determined based on the background color of `image`.
    :param thickness: The thickness of the text. If not provided, it will be calculated based on the image size.
    :param background_color: The background color behind the text. It can be specified as a string ("auto") or as a tuple of RGB values. If "auto" is specified,
                             the background color will be automatically determined based on the background color of `image`.
                             If a tuple is provided, the background color will be filled with that color.
    :param line_type: The type of line used for drawing the text. It should be one of the predefined constants from the cv2 library: e.g `cv2.LINE_AA=16`
    :param padding_pixels: The number of pixels used for padding around the text. This can only be applied together with string `position` e.g. "upper_left_corner"
    :param deep_copy: If true, copy `image` before drawing text on it
    """

    # Checks - simple
    # ------------------------------------------------------
    checker.assert_types(
        [text, position, font, font_scale, color, thickness, background_color, line_type, padding_pixels],
        [str, (str, tuple), int, int, tuple, int, (str, tuple), int, int],
        ["text", "position", "font", "font_scale", "color", "thickness", "background_color", "line_type", "padding_pixels"],
        [0, 1, 0, 1, 1, 1, 1, 0, 0],
    )

    # Checks - image
    image_source = parsers.parse_image_as_uint8_rgb_numpy_array(image_source)
    checker.assert_valid_image(image_source)
    if deep_copy:
        image_source = copy.deepcopy(image_source)
    if (len(image_source.shape) == 3) and (image_source.shape[-1] == 1):
        warnings.warn("You have provided an greyscale image in the format (dim1 x dim2 x 1), it will be squeezed down to (dim1 x dim2)")
        image_source = image_source[:, :, 0:1].squeeze(2)

    # Checks - cv2 specific stuff and background_color
    # ------------------------------------------------------
    if font not in [0, 1, 2, 3, 4, 5, 6, 7]:
        raise ValueError(
            f"Unrecognized `{font=}`. As of writing this, there's 8 valid fonts in cv2. These are encoded with an enum spanning from 0-7 (both ends included).")
    if line_type not in [4, 8, 16]:
        raise ValueError(
            f"Unrecognized `{line_type=}`. As of writing this, there's 3 valid fonts in cv2. These are encoded with an enum with [LINE_4=4, LINE_8=8, LINE_AA=16]")
    if font_scale is not None:
        assert font_scale > 0, f"Expected font_scale > 0, but received `{font_scale=}`"
    if color is not None:
        assert (len(color) == 3) and all(0 <= c <= 255 for c in color), f"Expected `color` to contain 3 RGB values between 0-255, but received `{color=}`"
    if thickness is not None:
        assert thickness > 0, f"Expected thickness > 0, but received `{thickness=}`"
    if isinstance(background_color, tuple):
        assert (len(background_color) == 3) and all(0 <= c <= 255 for c in background_color), \
            f"Expected `background_color` to contain 3 RGB values between 0-255, but received `{background_color=}`"
    if isinstance(background_color, str):
        assert background_color == "auto", f"The only valid string-argument for background_color is 'auto', but received {background_color=}"

    # Checks - Text position
    # ------------------------------------------------------
    legal_positions = ["upper_left_corner", "upper_right_corner", "bottom_left_corner", "bottom_right_corner", "center"]
    if isinstance(position, str):
        assert position in legal_positions, f"Invalid `{position=}`. Legal values are: `{legal_positions}`"
    assert 0 <= padding_pixels <= min([c // 2 for c in image_source.shape[:2]]), \
        f"Expected 0 <= padding_pixels <= min_image_size/2, but received `{padding_pixels=}`"

    if isinstance(position, tuple):
        assert len(
            position) == 2, f"Invalid `{position=}`. Expected a tuple with (text_left_corner_x, text_left_corner_right)"
        assert 0 <= position[0] <= image_source.shape[
            0], f"Invalid `{position[0]=}`. The first coordinate (width) is outside the image"
        assert 0 <= position[1] <= image_source.shape[
            1], f"Invalid `{position[1]=}`. The second coordinate (height) is outside the image"

        if padding_pixels:
            warning_message = f"`{padding_pixels=}` will be ignored. It only works with string `position` e.g. `upper_left_corner`"
            warnings.warn(warning_message)

    # Assign font_scale and thickness dynamically
    # ------------------------------------------------------
    if font_scale is None:
        font_scale = max(1, round(max(image_source.shape[:2]) / 1000))
    if thickness is None:
        thickness = max(1, round(max(image_source.shape[:2]) / 1000))

    # Handle position conversion
    # ------------------------------------------------------
    image_height, image_width = image_source.shape[0], image_source.shape[1]
    (text_width, text_height), _ = cv2.getTextSize(text, font, font_scale, thickness)
    if isinstance(position, str):
        if position == "upper_left_corner":
            position = (padding_pixels, text_height + padding_pixels)
        elif position == "upper_right_corner":
            position = (image_width - (text_width + padding_pixels), text_height + padding_pixels)
        elif position == "bottom_left_corner":
            position = (padding_pixels, image_height - padding_pixels)
        elif position == "bottom_right_corner":
            position = (image_width - (text_width + padding_pixels), image_height - padding_pixels)
        elif position == "center":
            position = ((image_width - text_width) // 2, (image_height + text_height) // 2)
        else:
            raise RuntimeError("Shouldn't have gotten this far")

    # Handle text and background colors
    # ------------------------------------------------------

    # Calculate some shared parameters for color calculations
    if (color is None) or (background_color is not None):
        extra_padding = text_height // 8

        # Find text area in the image
        x1 = max(0, position[0] - extra_padding)
        x2 = min(image_width, position[0] + extra_padding + text_width)
        y1 = max(0, position[1] - extra_padding - text_height)
        y2 = min(image_height, position[1] + extra_padding)

        # Determine if the text should be dark or bright
        background_pixels_mean = image_source[y1:y2, x1:x2].mean()
        is_dark = (background_pixels_mean < 128)

    # Background color user defined
    if isinstance(background_color, tuple):
        # why `ascontiguousarray` -> see explanation at the bottom of function
        image_source = cv2.rectangle(np.ascontiguousarray(image_source, dtype=np.uint8), (x1, y1), (x2, y2), background_color, -1)
        background_pixels_mean = image_source[y1:y2, x1:x2].mean()  # Must be recalculated due to the drawn box
        is_dark = (background_pixels_mean < 128)

    # Text color
    if color is None:
        color = (200, 200, 200) if is_dark else (50, 50, 50)

    # Background color -auto
    if background_color == "auto":
        background_image = np.zeros(image_source[y1:y2, x1:x2].shape)
        if not is_dark:
            background_image = background_image + 255
        background_image = background_image.astype(np.uint8)
        image_source[y1:y2, x1:x2] = cv2.addWeighted(image_source[y1:y2, x1:x2], 0.40, background_image, 0.60, 0.0)

    # Draw text to image
    # ------------------------------------------------------
    try:
        image_source = cv2.putText(image_source, text, position, font, font_scale, color, thickness, line_type)
    except Exception:
        # TODO: Find out if this is still an issue. I encountered a weird cv2 bug in which it refused
        # to acknowledge `image` as a valid np.array unless `np.ascontiguousarray` was applied. This fix
        # was mentioned in an official bug report on cv2's github page. But it should not be necessary.
        image_source = np.ascontiguousarray(image_source, dtype=np.uint8)
        image_source = cv2.putText(image_source, text, position, font, font_scale, color, thickness, line_type)

    return image_source

class ImageModifier:
    # TODO
    """
    Grayscale Conversion: Convert a color image to grayscale by averaging the RGB channels.
    Grayscale 2 RGB: Simply duplicate x 3
    Gamma Correction: Adjust the image's brightness by applying a power law transformation.
    Histogram Equalization: Enhance the contrast of an image by redistributing the intensity values using the image's histogram.
    Histogram Stretching: Expand the range of pixel intensities to enhance the contrast of an image.
    Image Rotation: Rotate the image by a specified angle.
    Image Flipping: Flip the image horizontally or vertically.
    Image Cropping: Crop a region of interest from the image.
    Image Resizing: Resize the image to a specified width and height.
    Image Padding: Add padding to the image by extending the border.
    Image Blurring: Apply blurring filters to the image to reduce noise or smooth out details.
    Image Sharpening: Enhance the image's details and edges.
    Image Thresholding: Convert a grayscale image to binary by applying a threshold value.
    Image Filtering: Apply various image filters like Gaussian blur, median filter, or custom filters.
    Image Edge Detection: Detect edges in the image using techniques like Canny edge detection or Sobel operator.
    Image Morphological Operations: Perform operations like dilation, erosion, opening, and closing on the image.
    Image Color Adjustment: Adjust the image's color balance, saturation, or hue.
    """


# noinspection PyTypeChecker
def draw_text_pillow(
        image_source:ImageType,
        x:int,
        y:int,
        text:str,
        font_size:int=20,
        font_thickness:int=4,
        color:Tuple[int, int, int]=(25,25,25),
        background_color:tuple=(240,240,240),
        italic:bool=False
) -> ndarray:
    """
    Draw `text` on `image_source` at the specified position (x, y) using PIL.

    :param image_source: An image. Can handle most image sources: ndarray, url, path, torch, PIL. For a full description see: TODO insert link
    :param x: The x-coordinate for the starting position of the text.
    :param y: The y-coordinate for the starting position of the text.
    :param text: The text to be drawn on the image.
    :param font_size: The font size of the text.
    :param font_thickness: The thickness of the text stroke.
    :param color: The color of the text in RGB format.
    :param background_color: The background color of the text box in RGB format.
    :param italic: Flag to make the text italic.

    :return: The modified image with the text drawn.
    """

    # Checks
    checker.assert_types(
        [x,   y,   text, font_size, font_thickness, color,  background_color, italic],
        [int, int, str,  int,       int,            tuple,  tuple,            bool],
        ["x", "y", "text", "font_size", "font_thickness", "color", "background_color", "italic"],
        [0,   0,   0,    0,         0,              0,      1,                0]
    )

    image_source = parsers.parse_image_as_uint8_rgb_numpy_array(image_source)
    checker.assert_positive_int(x, zero_allowed=True, max_value_allowed=image_source.shape[1], variable_name="x")
    checker.assert_positive_int(y, zero_allowed=True, max_value_allowed=image_source.shape[0], variable_name="y")
    checker.assert_color(color, color_type="RGB", variable_name="color")
    if background_color is not None:
        checker.assert_color(background_color, color_type="RGB", variable_name="background_color")

    # Draw
    font = utils.get_font(font_size, font_thickness, italic)
    image_source = Image.fromarray(image_source)
    draw = ImageDraw.Draw(image_source)
    if background_color:
        left, top, right, bottom = draw.textbbox((x,y ), text, font=font)
        draw.rectangle((left-4, top-2, right+4, bottom+2), fill=background_color)
    draw.text((x, y), text, color, font=font)
    return np.array(image_source)


# noinspection PyTypeChecker
def draw_image_title(
        image_source:ImageType,
        text:str,
        placement:str= "center",
        font_size:int=30,
        font_thickness:int=4,
        text_color:Tuple[int, int, int]=(25, 25, 25),
        text_background_color:tuple=None,
        top_padding:int=None,
        top_padding_color:Tuple[int, int, int]=(255, 255, 255),
        italic:bool=False
) -> ndarray:
    """
    Draw a title `text` on `image` using PIL.

    :param image_source: An image. Can handle most image sources: ndarray, url, path, torch, PIL. For a full description see: TODO insert link
    :param text: The title text to be drawn on the image.
    :param placement: The placement of the text ('center', 'left', 'right').
    :param font_size: The font size of the text.
    :param font_thickness: The thickness of the text stroke (1,2,3,4,5,6,7).
    :param text_color: The color of the text in RGB format.
    :param text_background_color: The background color of the text box in RGB format.
    :param top_padding: The additional padding at the top of the image (if required).
    :param top_padding_color: The color of the top padding.
    :param italic: Flag to make the text italic.
    """

    # Checks
    checker.assert_types(
        [text, placement, font_size, font_thickness, top_padding, italic],
        [str, str, int, int, int, bool],
        ["text", "placement", "font_size", "font_thickness", "top_padding", "italic"],
        [0, 0, 0, 0, 1, 0]
    )

    image_source = parsers.parse_image_as_uint8_rgb_numpy_array(image_source)
    checker.assert_color(text_color, color_type="RGB", variable_name="text_color")
    if text_background_color is not None:
        checker.assert_color(text_background_color, color_type="RGB", variable_name="text_background_color")
    if top_padding_color is not None:
        checker.assert_color(top_padding_color, color_type="RGB", variable_name="top_padding_color")
    checker.assert_in(placement, ["center", "left", "right"], "placement")

    # Setup
    image_height, image_width, _ = image_source.shape
    font = utils.get_font(font_size, font_thickness, italic)

    # Determine text box dimensions
    image_source = Image.fromarray(image_source)
    left, top, right, bottom = ImageDraw.Draw(image_source).textbbox((0, 0), text, font=font)
    text_box_height = bottom - top; assert text_box_height > 0
    text_box_width = right - left; assert text_box_width > 0

    # Padding
    if top_padding is None:
        top_padding = int(text_box_height * 2.5)
    image_source = cv2.copyMakeBorder(np.array(image_source), top_padding, 0, 0, 0, cv2.BORDER_CONSTANT, None, value=top_padding_color)

    # Determine text placement
    y = top_padding // 2 - text_box_height // 2 - top  # `top` is the font's built-in "white space" on top.
    if placement == "center":
        x = (image_width - text_box_width) // 2
    elif placement == "left":
        x = 10
    elif placement == "right":
        x = image_width - text_box_width - 10
    else:
        raise ValueError("Shouldn't have gotten this far!")

    assert y > 0
    if x < 0:
        msg = f"`{text=}` with `{placement=}` exceeds the image's border, and as a result, parts of the text will not be visible."
        warnings.warn(msg)

    # Draw title
    image_source = Image.fromarray(image_source)
    draw = ImageDraw.Draw(image_source)
    if text_background_color:
        left, top, right, bottom = draw.textbbox((x, y), text, font=font)
        draw.rectangle((left - 4, top - 2, right + 4, bottom + 2), fill=text_background_color)
    draw.text((x, y), text, text_color, font=font)
    return np.array(image_source)


def add_square_dropdown_shadow(image, shadow_upper_left_corner_xy, shadow_lower_right_corner_xy, shadow_offset_x, shadow_offset_y, shadow_color, background_color, iterations, blur_kernel):
    x1, y1 = shadow_upper_left_corner_xy
    x2, y2 = shadow_lower_right_corner_xy
    image[y1+shadow_offset_y:y2+shadow_offset_y, x1+shadow_offset_x:x2+shadow_offset_x, :] = shadow_color
    for _ in range(iterations):
        image = cv2.GaussianBlur(image, blur_kernel, 0)
    image[y1:y2, x1:x2, :] = background_color
    return image


__all__ = ["draw_text_cv2", "draw_text_pillow", "draw_image_title"]


from __future__ import annotations
import copy
import numpy as np
import cv2
import itertools
from PIL import Image, ImageDraw

from config import *
import show
import checker as checker
import image_modifier
import parsers
import utils

def _get_grid_dimensions(images:List[np.ndarray], rows:int, cols:int):
    # TODO: add return types
    coordinates = list(itertools.product(range(rows), range(cols)))
    coordinates = coordinates[:len(images)]
    grid_dims = np.zeros((rows, cols, 2))

    # Extract grid data
    for i, (row, col) in enumerate(coordinates):
        image_height, image_width = images[i].shape[:2]

        # Max col/row shape
        grid_dims[row, col, 0] = image_width
        grid_dims[row, col, 1] = image_height

    # Find the per columns/row maximum width/height
    max_height_per_row = grid_dims[:, :, 1].max(axis=1)
    assert len(max_height_per_row) == rows
    max_width_per_col = grid_dims[:, :, 0].max(axis=0)
    assert len(max_width_per_col) == cols

    return coordinates, max_width_per_col, max_height_per_row

def _get_canvas(dummy_draw, row_text, col_text, row_font, col_font, max_height_per_row, rows, row_spacing, text_adjusted_font_size, max_width_per_col, cols, col_spacing, canvas_background_color, image_area_margins):
    # Find out how much extra width/height is needed for row/col text
    extra_height_col_text, extra_width_row_text = 0, 0
    if row_text:
        extra_width_row_text =  int(max([dummy_draw.textbbox((0, 0), text, font=row_font)[2] for text in row_text]) * 1.2)
    if col_text:
        extra_height_col_text = int(max([dummy_draw.textbbox((0, 0), text, font=col_font)[3] for text in col_text]) * 1.2)

    # Margins
    margins_width = image_area_margins["left"] + image_area_margins["right"]
    margins_height = image_area_margins["top"] + image_area_margins["bottom"]

    # Construct the canvas find a suitable size given the inputs
    canvas_height = margins_height + extra_height_col_text + sum(max_height_per_row) + (rows-1)*row_spacing + (rows-1)*text_adjusted_font_size
    canvas_width =  margins_width  + extra_width_row_text  + sum(max_width_per_col)  + (cols-1)*col_spacing
    canvas_height, canvas_width = int(canvas_height), int(canvas_width)
    canvas = np.zeros((canvas_height, canvas_width, 3)).astype(np.uint8)
    canvas[...] = canvas_background_color
    return canvas, extra_height_col_text, extra_width_row_text


# noinspection PyTypeChecker
def _draw_text(canvas, x, y, text, color, font):
    canvas = Image.fromarray(canvas)
    draw = ImageDraw.Draw(canvas)
    draw.text((x, y), text, color, font=font)
    canvas = np.array(canvas)
    return canvas

def _get_cell_data(images, image_text, offset_x, offset_y, coordinates, dummy_draw, text_font, col_spacing, max_width_per_col, row_spacing, text_adjusted_font_size, max_height_per_row):
    cell_data = []

    for i, (row, col) in enumerate(coordinates):
        image = images[i]
        image_height, image_width, _ = image.shape
        cell_text, text_width, text_height = None, 0, 0
        if image_text is not None:
            cell_text = image_text[i]
            _, _, text_width, text_height = dummy_draw.textbbox((0, 0), cell_text, font=text_font)

        # Cell coordinates
        cell_upper_left_x = offset_x + sum(max_width_per_col[:col])  + col*col_spacing
        cell_upper_left_y = offset_y + sum(max_height_per_row[:row]) + row*(row_spacing+text_adjusted_font_size)
        cell_lower_right_x = cell_upper_left_x + max_width_per_col[col]
        cell_lower_right_y = cell_upper_left_y + max_height_per_row[row]

        # Image coordinates (centered within the cell width)
        image_upper_left_x = cell_upper_left_x + (max_width_per_col[col] - image_width)/2
        image_upper_left_y = cell_upper_left_y
        image_lower_right_x = image_upper_left_x + image_width
        image_lower_right_y = image_upper_left_y + image_height

        # Center of the cell
        cell_center_x = cell_upper_left_x + (cell_lower_right_x - cell_upper_left_x)/2
        cell_center_y = cell_upper_left_y + (cell_lower_right_y - cell_upper_left_y)/2

        assert int(image_lower_right_x - image_upper_left_x) == image_width,  "The calculated coordinates doesn't match with image's width"
        assert int(image_lower_right_y - image_upper_left_y) == image_height, "The calculated coordinates doesn't match with image's height"

        cell_data.append({
            "row":row,
            "col":col,
            "image": image,
            "image_height":image_height,
            "image_width":image_width,
            "cell_text":cell_text,
            "text_width":text_width,
            "text_height":text_height,
            "cell_upper_left_x":  int(cell_upper_left_x),
            "cell_upper_left_y":  int(cell_upper_left_y),
            "cell_lower_right_x":  int(cell_lower_right_x),
            "cell_lower_right_y":  int(cell_lower_right_y),
            "image_upper_left_x": int(image_upper_left_x),
            "image_upper_left_y": int(image_upper_left_y),
            "image_lower_right_x": int(image_lower_right_x),
            "image_lower_right_y": int(image_lower_right_y),
            "cell_center_x":      int(cell_center_x),
            "cell_center_y":      int(cell_center_y),
        })
    return cell_data

def _draw_separation_lines(canvas, sep_line_color, image_area_offset_y, rows, row_spacing, text_adjusted_font_size, max_height_per_row, image_area_offset_x, cols, col_spacing, max_width_per_col, cell_data, col_sep_line, row_sep_line):
    # If the line is this faint, I might as well just skip the line drawing all together, right?
    if sep_line_color[3] < 0.01:
        return canvas

    alpha_overlay = canvas.copy() if (sep_line_color and (sep_line_color[3] < 1.0)) else None
    y_min = int(image_area_offset_y)
    y_max = int(image_area_offset_y + (rows - 1) * (row_spacing + text_adjusted_font_size) + max_height_per_row.sum())
    x_min = int(image_area_offset_x)
    x_max = int(image_area_offset_x + (cols - 1) * col_spacing + max_width_per_col.sum())

    for cell in cell_data:
        x2, y2 = cell["cell_lower_right_x"], cell["cell_lower_right_y"]

        if (cell["row"] == 0) and (cell["col"] < (cols - 1)) and (col_sep_line > 0):
            x = round(x2 + col_spacing / 2 - round(col_sep_line / 2))  # The double round just looks better for whatever reason
            cv2.line(canvas, (x, y_min), (x, y_max), sep_line_color, col_sep_line)

        if (cell["col"] == 0) and (cell["row"] < (rows - 1)) and (row_sep_line > 0):
            y = round(y2 + row_spacing / 2 - round(row_sep_line / 2) + text_adjusted_font_size / 2)  # The double round just looks better for whatever reason
            cv2.line(canvas, (x_min, y), (x_max, y), sep_line_color, row_sep_line)

    # Line transparency
    if alpha_overlay is not None:
        alpha = sep_line_color[3]
        canvas = cv2.addWeighted(alpha_overlay, 1 - alpha, canvas, alpha, 0)

    return canvas

def _resize_images(images: List[np.ndarray], config: Dict):
    """
    So this is both inefficient and kinda ugly, and I'm unapologetic about it :)
    The thing is, Firstly, I don't want to have to write however many combinations there is for different values of
    [min_width, min_height, max_width, max_height]
    Secondly, I would really like to avoid messing about with the aspect ratio if at all possible.
    So I first attempt a clumsy, but aspect respecting resize and fall back on brute force resize if that fails
    """

    # Just return the images as is if nothing needs to be done
    if not any([config["max_width"], config["max_height"], config["min_width"], config["min_height"]]):
        return images

    # Setup
    max_width = config["max_width"] if config["max_width"] else 1e6
    max_height = config["max_height"] if config["max_height"] else 1e6
    min_width = config["min_width"]
    min_height = config["min_height"]
    interpolation = config["interpolation"]

    # Resize all images
    for i, image in enumerate(images):
        h, w, _ = image.shape

        # Attempt resize without altering the aspect ratio
        if min_width and (min_width > w):
            image, h, w = image_modifier.resize_image_respect_aspect_ratio(image, min_width, None, interpolation)
        if min_height and (min_height > h):
            image, h, w = image_modifier.resize_image_respect_aspect_ratio(image, None, min_height, interpolation)
        if max_width and (max_width < w):
            image, h, w = image_modifier.resize_image_respect_aspect_ratio(image, max_width, None, interpolation)
        if max_height and (max_height < h):
            image, h, w = image_modifier.resize_image_respect_aspect_ratio(image, None, max_height, interpolation)

        # If all else fail, ignore the aspect ratio and ensure the resize respect the users input
        if (h < min_height) or (h > max_height) or (w < min_width) or (w > max_width):
            width = min_width if (w < min_width) else min(w, max_width)
            height = min_height if (h < min_height) else min(h, max_height)
            image = cv2.resize(image, (width, height), interpolation=interpolation)
        images[i] = image
    return images

def _add_padding(images, border_size, color, add_padding_on_the_outside):
    if not border_size:
        return images

    s = border_size
    for i, image in enumerate(images):
        if add_padding_on_the_outside:
            padded_image = cv2.copyMakeBorder(image, s, s, s, s, cv2.BORDER_CONSTANT, value=color)
        else:
            padded_image = np.full(image.shape, color, dtype=np.uint8)
            padded_image[s:-s, s:-s, :] = image[s:-s, s:-s, :]
        images[i] = padded_image
    return images

def _check_dict(image_spacing:dict, image_resizing:dict, image_text_config:dict, col_text_config:dict, row_text_config:dict, image_area_margins:dict, image_border:dict, drop_down_shadow:dict):

    # --------------------------------------------------------------------
    # Checks - `image_spacing`
    # --------------------------------------------------------------------

    if image_spacing.get("sep_line_color"):
        checker.assert_color(image_spacing["sep_line_color"], "RGBA", 'image_spacing["sep_line_color"]')

    checker.assert_dict_types(
        dict_name="image_spacing",
        dict_to_check=image_spacing,
        names=["col_spacing", "row_spacing", "col_sep_line", "row_sep_line", "sep_line_color"],
        types=[int, int, int, int, tuple],
        must_be_present=[1, 1, 1, 1, 1],
        allow_nones=[0, 0, 0, 0, 0],
        eval_expr=["0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6", None],
        assert_dict_lte=True
    )

    # `col_spacing`, `row_spacing`
    checker.assert_eval(image_spacing["col_spacing"], "0<=x", False, "col_spacing")
    checker.assert_eval(image_spacing["row_spacing"], "0<=x", False, "row_spacing")

    # `col_sep_line`, `row_sep_line`
    if image_spacing["col_sep_line"]:
        assert image_spacing["col_sep_line"] < (image_spacing["row_spacing"] + 2)
    if image_spacing["row_sep_line"]:
        assert image_spacing["row_sep_line"] < (image_spacing["col_spacing"] + 2)

    # `canvas_background_color`, `sep_line_color`
    if image_spacing["sep_line_color"]:
        checker.assert_color(image_spacing["sep_line_color"], "RGBA", 'image_spacing["sep_line_color"]')

    # --------------------------------------------------------------------
    # Checks - `image_resize`
    # --------------------------------------------------------------------

    if image_resizing is not None:
        if image_resizing.get("interpolation") and not (0 <= image_resizing["interpolation"] <= 7):
            raise ValueError("As of writing this, there's 8 valid interpolation in cv2. These are encoded with an enum spanning from 0-7 (both ends included)."
                             "And include the following names (`cv2.` prefix):"
                             "[INTER_NEAREST, INTER_LINEAR, INTER_LINEAR_EXACT, INTER_AREA, INTER_CUBIC, INTER_LANCZOS4, INTER_NEAREST_EXACT, INTER_MAX]")
        checker.assert_dict_types(
            dict_name="image_resizing",
            dict_to_check=image_resizing,
            names=["max_width", "max_height", "min_width", "min_height", "interpolation"],
            types=[int, int, int, int, int],
            must_be_present=[0, 0, 0, 0, 0],
            allow_nones=[1, 1, 0, 0, 0],
            eval_expr=["0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6", "0<=x<=7"],
            assert_dict_lte=True
        )

    # --------------------------------------------------------------------
    # Checks - `image_text_config`, `col_text_config`, `row_text_config`
    # --------------------------------------------------------------------

    # TODO: This is what I want to have implemented
    # to_check = [
    #     (image_text_config, "image_text_config", ['title_left', 'title_middle', 'left', 'middle']),
    #     (col_text_config, "col_text_config", ['title_left', 'title_middle', 'left', 'middle']),
    #     (row_text_config, "row_text_config", ['right'])
    # ]
    # This is what is actually implemented
    to_check = [
        (image_text_config, "image_text_config",['title_middle']),
        (col_text_config, "col_text_config",    ['title_middle']),
        (row_text_config, "row_text_config",    ['right'])
    ]

    for (dict_to_check, dict_name, legal_placements) in to_check:
        if dict_to_check is None:
            continue

        # Manual checks
        if dict_to_check.get("color"):
            checker.assert_color(dict_to_check["color"], "RGB", 'dict_to_check["color"]')
        if dict_to_check.get("placement"):
            checker.assert_in(dict_to_check["placement"], legal_placements, f"{dict_name}['placement']")
        if dict_to_check.get("font_thickness") is not None:
            if not (0<dict_to_check["font_thickness"]<=7):
                raise ValueError(f"`{dict_name}['font_thickness']={dict_to_check['font_thickness']}` is not valid. As of writing this, `font_thickness` must be in [1,2,3,4,5,6,7]")

        checker.assert_dict_types(
            dict_name=dict_name,
            dict_to_check=dict_to_check,
            names=['placement', 'font_size', 'font_thickness', 'italic', 'color', 'adjust_draw_distance'],
            types=[str, int, int, bool, tuple, (str, int)],
            must_be_present=[1, 1, 1, 1, 1, 1],
            allow_nones=[0, 0, 0, 0, 0, 1],
            eval_expr=[None, "0<=x<=1e6", "0<x<=7", None, None, None],
            assert_dict_lte=True
        )

    # --------------------------------------------------------------------
    # Checks - `image_area_margins`
    # --------------------------------------------------------------------

    if image_area_margins is not None:
        checker.assert_dict_types(
            dict_name="image_area_margins",
            dict_to_check=image_area_margins,
            names=["left", "right", "top", "bottom"],
            types=[int, int, int, int],
            must_be_present=[1, 1, 1, 1],
            allow_nones=[0, 0, 0, 0],
            eval_expr=["0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6", "0<=x<=1e6"],
            assert_dict_lte=True
        )

    # --------------------------------------------------------------------
    # Checks - `image_border`
    # --------------------------------------------------------------------

    if image_border is not None:
        if image_border.get("color"):
            checker.assert_color(image_border["color"], "RGB", 'image_border["color"]')
        checker.assert_dict_types(
            dict_name="image_border",
            dict_to_check=image_border,
            names=["border_size", "color", "add_padding_on_the_outside"],
            types=[int, tuple, int],
            must_be_present=[1, 1, 1],
            allow_nones=[0, 0, 0],
            eval_expr=["0<=x<=1e6", None, None],
            assert_dict_lte=True
        )

    # --------------------------------------------------------------------
    # Checks - `drop_down_shadow`
    # --------------------------------------------------------------------

    if drop_down_shadow is not None:
        if drop_down_shadow.get("color"):
            checker.assert_color(drop_down_shadow["shadow_color"], "RGB", 'drop_down_shadow["shadow_color"]')
        checker.assert_dict_types(
            dict_name="drop_down_shadow",
            dict_to_check=drop_down_shadow,
            names=["shadow_offset_x", "shadow_offset_y", "shadow_color", "iterations", "blur_kernel", "shadow_margins_lrtb"],
            types=[int, int, tuple, int, tuple, tuple],
            must_be_present=[1, 1, 1, 1, 1, 1],
            allow_nones=[0, 0, 0, 0, 0, 0],
            eval_expr=["x<=1e6", "x<=1e6", None, "0<=x<=1e6", "len(x)==2", "len(x)==4"],
            assert_dict_lte=True
        )
        if any([((k % 2 == 0) or (0 > k)) for k in drop_down_shadow["blur_kernel"]]):
            raise ValueError(f"`{drop_down_shadow['blur_kernel']=}` is not valid. It should contain exactly 2 odd, strictly positive numbers")
        if drop_down_shadow["iterations"] > 10:
            msg = f"{drop_down_shadow['iterations']=} seems unnecessarily high. The blur effect is pretty expensive."
            warnings.warn(msg)
        if any(0 > n for n in drop_down_shadow["shadow_margins_lrtb"]):
            raise ValueError(f"`{drop_down_shadow['shadow_margins_lrtb']=}` is not valid. All margins should be non-negative")

def _attempt_to_determine_rows_and_or_cols(rows, row_text, cols, col_text, num_images):
    if (row_text is not None) and (rows is None):
        raise ValueError("You have provided `row_text` but not `rows`. Please provide both or none.")
    if (col_text is not None) and (cols is None):
        raise ValueError("You have provided `col_text` but not `cols`. Please provide both or none.")

    if (rows is None) and (cols is None):
        for rows in range(100):
            cols = int(rows * (16/9))
            if (rows * cols) >= num_images:
                break
        if ((rows-1) * cols) >= num_images:
            rows = rows - 1
    elif (rows is None) and (cols is not None):
        rows = int(np.ceil(num_images/cols))
    elif (rows is not None) and (cols is None):
        cols = int(np.ceil(num_images / rows))

    failed_auto_detect = (not cols) or (not rows) or (rows*cols < num_images)
    if failed_auto_detect:
        raise ValueError("Failed to automatically determine `rows` and `cols` please provide both manually")
    return rows, cols

def _get_font_grid(text, placement: str, font_size: int, font_thickness: int, italic: bool, color, adjust_draw_distance: int):
    if text is None:
        return None, 0

    font = utils.get_font(font_size, font_thickness, italic)

    # Adjusted font size
    adjusted_font_size = font_size if ("title" in placement) else 0
    if adjust_draw_distance is None:
        adjusted_font_size += max(5, int(font_size / 3))
    else:
        adjusted_font_size += adjust_draw_distance
    return font, adjusted_font_size

# noinspection PyDefaultArgument
def show_grid_configurable(
        image_source: ImageSource,
        rows: int = None,
        cols: int = None,
        image_text: List[str] = None,
        col_text: List[str] = None,
        row_text: List[str] = None,
        canvas_background_color: Tuple[int, int, int] = (255, 255, 255),
        title: Optional[str] = None,
        display_image: bool = True,
        return_image: bool = False,
        max_output_image_size_wh: Optional[tuple] = (1920, 1080),
        save_image_path: Optional[str] = None,
        resize_factor: float = 1.0,
        BGR2RGB: bool = False,
        parse_image: bool = True,
        c_image_spacing:      Optional[Dict] = {"col_spacing":5, "row_spacing":5, "col_sep_line":1, "row_sep_line":1, "sep_line_color":(175, 175, 175, 0.3)},
        c_image_area_margins: Optional[Dict] = {"left": 40, "right":40, "top":40, "bottom":40},
        c_image_resizing:     Optional[Dict] = None,
        c_image_text_config:  Optional[Dict] = {'placement': 'title_middle', 'font_size': 15, 'font_thickness': 3, 'italic': False, 'color': (75, 75, 75), 'adjust_draw_distance': None},
        c_col_text_config:    Optional[Dict] = {"placement": "title_middle", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance":40},
        c_row_text_config:    Optional[Dict] = {"placement": "right", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance":40},
        c_image_border:       Optional[Dict] = None,
        c_drop_down_shadow:   Optional[Dict] = None,
) -> Optional[ndarray]:
    # TODO: Implement the different placements. (DONE, but not tested) Or at a minimum throw a NotImplementedError
    """

    Display a grid of images from `image_source` with specified number of `rows` and `cols`.
    NOTE: arguments prefixed with `c_` are dicts with extra configuration details. See examples below

    :param image_source: Two or more images. Can handle most image sources: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param rows: The number of rows in the grid. If not specified, it is calculated automatically.
    :param cols: The number of columns in the grid. If not specified, it is calculated automatically.
    :param image_text: List of texts to be displayed above each image.
    :param col_text: List of texts to be displayed at the top of each column.
    :param row_text: List of texts to be displayed on the right side of each row.
    :param canvas_background_color: Background color of the canvas in RGB format.
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.
    :param parse_image: If True, will thoroughly check image_source and ensure it's in the correct format
    :param c_image_spacing: Configuration for image spacing, margins, and separation lines. If None, nothing will be applied.
    Examples: {"col_spacing":40, "row_spacing":30, "col_sep_line":1, "row_sep_line":1, "sep_line_color":(175, 175, 175, 0.3)}
    :param c_image_area_margins: Configuration for image area margins. If None, nothing will be applied.
    Example: {"left": 40, "right":40, "top":40, "bottom":40}`
    :param c_image_resizing: Configuration for image resizing. If None, nothing will be applied.
    Example: {"max_width":800, "max_height":500, "min_width":0, "min_height":0, "interpolation":cv2.INTER_CUBIC}
    :param c_image_text_config: Configuration for image text. If None, nothing will be applied.
    Example: {"placement": "title_middle", "font_size": 15, "font_thickness": 3, "italic": False, "color": (75, 75, 75), "adjust_draw_distance":None}
    :param c_col_text_config: Configuration for column text. If None, nothing will be applied.
    Example: {"placement": "title_middle", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance":40}
    :param c_row_text_config: Configuration for row text. If None, nothing will be applied.
    Example: {"placement": "right", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance":40}
    :param c_image_border: Configuration for image border. If None, nothing will be applied.
    Example: {"border_size":1, "color":(200, 200, 200), "add_padding_on_the_outside":False}
    :param c_drop_down_shadow: Configuration for drop-down shadow. If None, nothing will be applied.
    :returns: If return_image is True, return an image of shape np.ndarray (H, W, 3), otherwise None
    """

    # --------------------------------------------------------------------
    # # Universal stuff and image parsing
    # --------------------------------------------------------------------

    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image)

    not_iterable = not isinstance(image_source, (tuple, list))
    is_tensor = (TORCH_FOUND and isinstance(image_source, torch.Tensor))
    is_numpy = isinstance(image_source, np.ndarray)
    general_invalid_msg = "Please provide one of the following:\n" \
                          "(1) a batch np.ndarray images (N, height, width, 3), a torch.Tensor batch or\n" \
                          "(2) a batch of torch.Tensor images (N, 3, height, width)\n" \
                          "(3) a list of images comprised of (np.ndarray, torch.Tensor, PIL.Image.Image, url:str, path:str)"
    if (not_iterable and (not is_tensor) and (not is_numpy)) or (image_source is None):
        raise ValueError(f"`{type(image_source)=}` is not valid."+general_invalid_msg)
    elif not_iterable: # TODO test batch loading
        if (is_numpy or is_tensor) and (not parse_image):
            msg = f"`{parse_image=}` can only be used together with a iterable `type(image_source)`. Will continue parse_image=False"
            warnings.warn(msg)
        if isinstance(image_source, torch.Tensor) and checker.assumed_torch_batch(image_source):
            image_source = parsers.parse_torch_image_batch_as_uint8_rgb_numpy_array(image_source, resize_factor, BGR2RGB)
        elif isinstance(image_source, np.ndarray) and checker.assumed_numpy_batch(image_source):
            image_source = parsers.parse_numpy_image_batch_as_uint8_rgb_numpy_array(image_source, resize_factor, BGR2RGB)
        else:
            raise ValueError(f"You have provided a `{type(image_source)=}` but could not parse it as a batch of images."+general_invalid_msg)
    else:
        if parse_image:
            for i, image in enumerate(image_source):
                    image_source[i] = parsers.parse_image_as_uint8_rgb_numpy_array(image, resize_factor, BGR2RGB)
        else:
            image_source = parsers.quick_parse_image_source(image_source)

    images = image_source # A bit cumbersome to carry around, but I wanted to keep `image_source` for the sake of consistency between show-functions
    assert len(images) >= 2, f"Expected at least two images, but received `{len(images)=}`"

    max_amount_reached = (len(images) > global_config.show_max_image_amount)
    if max_amount_reached:
        if (rows is None) and (cols is None):
            images = parsers.threshold_image_count(images)
        else:
            msg = f"`{len(images)=}` exceeds the limit `{global_config.show_max_image_amount}`, but the program will continue as is because rows/cols are specified."
            warnings.warn(msg)


    # --------------------------------------------------------------------
    # Checks - Types
    # --------------------------------------------------------------------

    # Type checks
    checker.assert_types(
        [rows,   cols,   image_text,   col_text,   row_text,     canvas_background_color,   c_image_spacing,   c_image_resizing, c_image_text_config, c_col_text_config, c_row_text_config, c_image_area_margins, c_image_border, c_drop_down_shadow],
        [int,    int,    list,         list,       list,         tuple,                     dict,              dict,             dict,                dict,              dict,              dict,                 dict,           dict],
        ["rows", "cols", "image_text", "col_text", "row_text",   "canvas_background_color", "c_image_spacing", "image_resizing", "image_text_config", "col_text_config", "row_text_config", "image_area_margins", "image_border", "drop_down_shadow"],
        [True,   True,   True,         True,       True,         False,                     True,              True,             True,                True,              True,              True,                 True,           True]
    )

    # List type checks
    if image_text: checker.assert_list_types(image_text, str, "image_text")
    if col_text: checker.assert_list_types(col_text, str, "col_text")
    if row_text: checker.assert_list_types(row_text, str, "row_text")

    # --------------------------------------------------------------------
    # Copy mutable default arguments to avoid any monkey business
    # --------------------------------------------------------------------
    # Note: Mutable defaults are chosen because the user can directly see what options can be used

    c_image_spacing = copy.deepcopy(c_image_spacing)
    c_image_area_margins = copy.deepcopy(c_image_area_margins)
    c_image_resizing = copy.deepcopy(c_image_resizing)
    c_image_text_config = copy.deepcopy(c_image_text_config)
    c_col_text_config = copy.deepcopy(c_col_text_config)
    c_row_text_config = copy.deepcopy(c_row_text_config)
    c_image_border = copy.deepcopy(c_image_border)
    c_drop_down_shadow = copy.deepcopy(c_drop_down_shadow)

    # --------------------------------------------------------------------
    # Assign default values if necessary
    # --------------------------------------------------------------------

    # Text will be set to zero height
    if image_text is None:
        text_adjusted_font_size = 0

    # Margins will all be set to zero
    if c_image_area_margins is None:
        c_image_area_margins = {"left": 0, "right": 0, "top": 0, "bottom": 0}

    # All spacing will be set to zero if None
    default_image_spacing = {"col_spacing": 0, "row_spacing": 0, "col_sep_line": 0, "row_sep_line": 0, "sep_line_color": (0, 0, 0, 1.0)}
    if c_image_spacing is None:
        c_image_spacing = default_image_spacing

    # If only some of the resize parameters are specified, default the rest
    if (c_image_resizing is not None) and (len(c_image_resizing) != 5):
        image_resizing_temp = {"max_width":int(1e6), "max_height":int(1e6), "min_width":0, "min_height":0, "interpolation":cv2.INTER_CUBIC}
        for k, v in c_image_resizing.items(): image_resizing_temp[k]=v
        c_image_resizing = image_resizing_temp; del image_resizing_temp

    # If the user has defined some variable, but not all, the remaining ones will get assigned default values
    for k, v in c_image_spacing.items():
        if c_image_spacing.get(k) is None:
            c_image_spacing[k] = default_image_spacing[k]

    # If the user haven't specified both `cols` and `rows` they will be chosen automatically
    if (rows is None) or (cols is None):
        rows, cols = _attempt_to_determine_rows_and_or_cols(rows, row_text, cols, col_text, num_images=len(images))

    # --------------------------------------------------------------------
    # Checks - Values
    # --------------------------------------------------------------------

    checker.assert_color(canvas_background_color, "RGB", "canvas_background_color")

    if not any([display_image, return_image, save_image_path]):
        raise ValueError("If you don't want to display, return or save the image. What's the point :)")

    # Dict checks (this is primarily type checks, but also some value checks)
    _check_dict(c_image_spacing, c_image_resizing, c_image_text_config, c_col_text_config, c_row_text_config, c_image_area_margins, c_image_border, c_drop_down_shadow)

    # These variables are littered all over the place, so I don't want to have to access them through a dict everytime
    col_spacing = c_image_spacing["col_spacing"]
    row_spacing = c_image_spacing["row_spacing"]
    col_sep_line = c_image_spacing["col_sep_line"]
    row_sep_line = c_image_spacing["row_sep_line"]
    sep_line_color = c_image_spacing["sep_line_color"]

    # `rows`, `cols`
    assert (rows > 0) and (cols > 0), f"Both `{rows=}` and `{cols=}` must be greater then 1"
    assert (rows * cols) >= len(images), f"You have provided too few `{cols=}` and/or `{rows=}`. `{rows*cols=}` is less then `{len(images)=}`"

    # `image_text`
    if image_text is not None:
        assert len(image_text) == len(images), f"Mismatch between the number of images `{len(images)=}` and the number of image texts `{len(image_text)=}`\n" \
                                               f"Provide a matching number or set `image_text=None`"
        if not all(len(t.splitlines()) <= 1 for t in image_text):
            msg = "Multiline strings are not supported because they cause all sorts of issues with the image layout." \
                  "\nChange the following entries in `image_text` to resolve the issue:\n"
            msg += ".\n".join([repr(t) for t in image_text if (len(t.splitlines()) != 1)])
            raise ValueError(msg)

    # `col_text` `row_text`
    if (col_text is not None) and (len(col_text) != cols):
        raise ValueError(f"There's a mismatch between the specified `{cols=}` and the number of `{len(col_text)=}`")
    if (row_text is not None) and (len(row_text) != rows):
        raise ValueError(f"There's a mismatch between the specified `{rows=}` and the number of `{len(row_text)=}`")

    #`image_text`, `image_text_config`, `col_text`, `col_text_config`, `row_text`, `row_text_config`
    if image_text and (not c_image_text_config):
        template = str({"placement": "title_middle", "font_size": 15, "font_thickness": 3, "italic": False, "color": (75, 75, 75), "adjust_draw_distance": None})
        raise ValueError(f"You must provide a valid `image_text_config` together with your `image_text`. Here a functional template:\nimage_text_config={template}")
    if col_text and (not c_col_text_config):
        template = str({"placement": "middle", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance": 70})
        raise ValueError(f"You must provide a valid `col_text_config` together with your `col_text`. Here a functional template:\ncol_text_config={template}")
    if row_text and (not c_row_text_config):
        template = str({"placement": "right", "font_size": 22, "font_thickness": 5, "italic": False, "color": (75, 75, 75), "adjust_draw_distance": 40})
        raise ValueError(f"You must provide a valid `row_text_config` together with your `row_text`. Here a functional template:\rows_text_config={template}")

    # `canvas_background_color`, `sep_line_color`
    checker.assert_color(canvas_background_color, "RGB", 'canvas_background_color')

    # --------------------------------------------------------------------
    # Initialization
    # --------------------------------------------------------------------

    # Text
    dummy_draw = ImageDraw.Draw(Image.new("RGB", (1, 1), (255, 255, 255)))
    text_font, col_font, row_font = None, None, None
    col_adjusted_font_size, row_adjusted_font_size = None, None
    if image_text:
        text_font, text_adjusted_font_size = _get_font_grid(image_text, **c_image_text_config)
    if col_text:
        col_font, col_adjusted_font_size = _get_font_grid(col_text, **c_col_text_config)
    if row_text:
        row_font, row_adjusted_font_size = _get_font_grid(row_text, **c_row_text_config)
    # TODO: Add warnings for if the text exceeds the maximum height and or width for it's column/row
    # TODO: Add warnings if the image_text exceeds the boundaries of its cell

    # Reformat, resize and padding
    if c_image_border:
        images = _add_padding(images, **c_image_border)
    if c_image_resizing:
        images = _resize_images(images, c_image_resizing)

    # Grid and canvas
    coordinates, max_width_per_col, max_height_per_row = _get_grid_dimensions(images, rows, cols)
    canvas, extra_height_col_text, extra_width_row_text = _get_canvas(dummy_draw, row_text, col_text, row_font, col_font, max_height_per_row, rows, row_spacing, text_adjusted_font_size, max_width_per_col, cols, col_spacing, canvas_background_color, c_image_area_margins)
    image_area_offset_x, image_area_offset_y = extra_width_row_text + c_image_area_margins["left"], extra_height_col_text + c_image_area_margins["top"]
    cell_data = _get_cell_data(images, image_text, image_area_offset_x, image_area_offset_y, coordinates, dummy_draw, text_font, col_spacing, max_width_per_col, row_spacing, text_adjusted_font_size, max_height_per_row)

    # --------------------------------------------------------------------
    # Draw image area
    # --------------------------------------------------------------------

    # Dropdown shadow
    if c_drop_down_shadow is not None:
        (left, right, top, bottom) = c_drop_down_shadow.pop("shadow_margins_lrtb")
        shadow_upper_left_corner_xy = (image_area_offset_x - left, image_area_offset_y - top)
        shadow_lower_right_corner_xy = (canvas.shape[1] - c_image_area_margins["right"] + right, canvas.shape[0] - c_image_area_margins["bottom"] + bottom)
        canvas = image_modifier.add_square_dropdown_shadow(canvas, shadow_upper_left_corner_xy, shadow_lower_right_corner_xy, background_color=canvas_background_color, **c_drop_down_shadow)
        # TODO add warning if the the shadow margins exceed the shape of the canvas

    # Draw separation lines between images
    if sep_line_color and (sep_line_color[3] < 1.0) and ((col_sep_line > 0) or (row_sep_line > 0)):
        canvas = _draw_separation_lines(canvas, sep_line_color, image_area_offset_y, rows, row_spacing, text_adjusted_font_size, max_height_per_row, image_area_offset_x, cols, col_spacing, max_width_per_col, cell_data, col_sep_line, row_sep_line)

    # Draw the content of each cell
    for cell in cell_data:
        x1, y1 = cell["image_upper_left_x"], cell["image_upper_left_y"]
        x2, y2 = cell["image_lower_right_x"], cell["image_lower_right_y"]
        canvas[y1:y2, x1:x2, :] = cell["image"] # Draw image to canvas

        # Draw cell text
        if cell["cell_text"] is not None:
            text_x = round(cell["cell_center_x"] - cell["text_width"]/2)
            text_y = round(y1 - text_adjusted_font_size)
            canvas = _draw_text(canvas, text_x, text_y, cell["cell_text"], c_image_text_config["color"], text_font)

    # --------------------------------------------------------------------
    # Draw column and row text
    # --------------------------------------------------------------------

    x_min, y_min = image_area_offset_x, image_area_offset_y
    for cell in cell_data:
        x_center, y_center = cell["cell_center_x"], cell["cell_center_y"]
        col, row = cell["col"], cell["row"]

        if (col_text is not None) and (cell["row"] == 0):
            text = col_text[col]
            text_width = dummy_draw.textbbox((0, 0), text, font=col_font)[2]
            x = round(x_center - text_width/2)
            y = round(y_min - col_adjusted_font_size)
            canvas = _draw_text(canvas, x, y, text, c_col_text_config["color"], col_font)

        if (row_text is not None) and (cell["col"] == 0):
            text = row_text[row]
            _, _, text_width, text_height = dummy_draw.textbbox((0, 0), text, font=row_font)
            x = round(x_min - row_adjusted_font_size - text_width)
            y = round(y_center - text_height)
            canvas = _draw_text(canvas, x, y, text, c_row_text_config["color"], row_font)

    # Universal stuff
    if title is not None:
        canvas = image_modifier.draw_image_title(canvas, title, top_padding_color=canvas_background_color)
    if max_output_image_size_wh is not None:
        canvas = image_modifier.resize_universal_output_image(canvas, max_output_image_size_wh)
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, canvas)
    if display_image:
        show.show_image(canvas)
    if return_image:
        return canvas


def get_all_possible_aspect_ratios(images, flatten:bool=True):
    # TODO: This function is used by `just_show` and should probably be cleaned up a bit
    def _remove_diagonal_elements(matrix: np.ndarray):
        assert len(matrix.shape) == 2
        n, m = matrix.shape
        assert n > 1

        # Get the indices for the diagonal elements
        diagonal_indices = np.arange(matrix.shape[0])

        # Create a boolean mask to exclude the diagonal elements
        mask = np.ones(matrix.shape, dtype=bool)
        mask[diagonal_indices, diagonal_indices] = False

        # Apply the mask to remove the diagonal elements
        result = matrix[mask].reshape(matrix.shape[0], -1)
        return result

    shapes = np.stack([image.shape for image in images])
    heights, widths, channels = shapes[:, 0], shapes[:, 1], shapes[:, 2]

    # Aspect ratio
    aspect_ratios = widths / heights
    ar_diffs = np.abs(aspect_ratios[:, np.newaxis] - aspect_ratios)  # abs(AR_n - AR_m) for n, m in [1, ..., len(aspect_ratios)]
    ar_diffs = _remove_diagonal_elements(ar_diffs)

    if flatten:
        ar_diffs = ar_diffs.flatten()
    return ar_diffs


def _get_grid_parameters(num_images, height, width, max_width=1920, max_height=1080, desired_ratio=9/17, ratio_weight=1.0, scale_weight=1.0, empty_weight=1.0):
    """
    Try at estimate #cols, #rows and resizing factor necessary for displaying a list of images in a visually pleasing way.
    This is essentially done by minimizing 3 separate parameters:
        (1) difference between `desired_ratio` and height/width of the final image
        (2) Amount of image scaling necessary
        (3) the number of empty cells (e.g. 3 images on a 2x2 --> empty_cell = 1)

    NOTE1: This was a pretty challenging function to write and the solution may appear a bit convoluted.
           I've included some notes at "doc/_get_grid_parameters.jpg" which will hopefully motivate the solution -
           in particular the loss-function used for optimization.
    NOTE2: This function is only intended to be used by `show()`

    @param max_height:
    @param max_width:
    @param desired_ratio:
    @return: cols, rows, scaling_factor, loss_info
    """

    assert num_images > 1
    assert (height >= global_config.image_min_height) and (width >= global_config.image_min_width)

    N = num_images
    h, w = height, width
    H, W = max_height, max_width

    losses = {}
    losses_split = []
    for a in [0.05 + 0.01 * i for i in range(96)]:
        for x in range(1, N + 1):
            for y in range(1, N + 1):

                # If the solution is not valid continue
                if (h * a * y > H) or (w * a * x > W) or (x * y < N):
                    continue
                # Otherwise calculate loss
                else:
                    ratio_loss = abs((h * y) / (w * x) - desired_ratio) # (1)
                    ratio_loss = ratio_weight * ratio_loss
                    scale_loss = (1 - a) ** 2 # (2)
                    scale_loss = scale_weight * scale_loss
                    empty_cell_loss = x*y/N - 1 # (3)
                    empty_cell_loss = empty_weight * empty_cell_loss
                    losses[(x, y, a)] = ratio_loss + scale_loss + empty_cell_loss
                    losses_split.append([ratio_loss, scale_loss, empty_cell_loss])

    if len(losses_split) == 0:
        raise RuntimeError("Failed to produce grid parameters. There can be several reasons for this, "
                           "but the most likely cause is having too many/large images")

    # TODO: Should this still be here? It was useful while debugging, but would this ever be used again?
    # if global_config.in_debug_mode:
    #     import pandas as pd
    #     df = pd.DataFrame([[a, b, c, d] for (a, b, c), d in losses.items()], columns=["y", "x", "a", "total"])
    #     df[["ratio_loss", "scale_loss", "empty_loss"]] = np.array(losses_split)
    #     global_config.df = df


    # pick parameters with the lowest loss
    rl, sl, ecl = losses_split[np.argmin(list(losses.values()))]
    loss_info = {"ratio":rl, "scale":sl, "empty_cell":ecl, "total":rl+sl+ecl}
    cols, rows, scaling_factor = min(losses, key=losses.get)
    return cols, rows, scaling_factor, loss_info


def show_grid(
    image_source:ImageSource,
    allow_auto_resize:bool=True,
    title: Optional[str] = None,
    display_image: bool = True,
    return_image: bool = False,
    max_output_image_size_wh: Optional[tuple] = (1920, 1080),
    save_image_path: Optional[str] = None,
    resize_factor: float = 1.0,
    BGR2RGB: bool = False,
    parse_image: bool = True,
) -> Optional[ndarray]:
    """
    Display `image_source` in an appropriate way, regardless of its format.

    :param image_source: An image or list of images. Can handle most image sources: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param allow_auto_resize: If set to True, the function can automatically resize images to achieve a more optimal layout.
           Note that the interaction between `allow_auto_resize=True` and `resize_factor != 1.0` can be somewhat unpredictable.
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.
    :param parse_image: If True, will thoroughly check image_source and ensure it's in the correct format

    """
    # --------------------------------------------------------------------
    # # Universal stuff
    # --------------------------------------------------------------------

    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image)
    if parse_image:
        image_source = parsers.parse_arbitrary_image_source(image_source, resize_factor, BGR2RGB)
    else:
        image_source = parsers.quick_parse_image_source(image_source)
    images = image_source  # A bit cumbersome to carry around, but I wanted to keep `image_source` for the sake of consistency between show-functions

    # If there's only one image just pass it to _ju
    assert len(images) >= 1, f"Expected at least one image, but received `{len(images)=}`"
    if len(images) == 1:
        return show.show_image(image_source=images[0], title=title, display_image=display_image, return_image=return_image,
                               max_output_image_size_wh=max_output_image_size_wh, save_image_path=save_image_path,
                               resize_factor=resize_factor, BGR2RGB=BGR2RGB, parse_image=False)

    # ------------------------------------------
    # Setup
    # ------------------------------------------

    # Image shape
    shapes = np.stack([image.shape for image in images])
    heights, widths, channels = shapes[:, 0], shapes[:, 1], shapes[:, 2]

    # Aspect ratio
    # If the difference in aspect ratios are less than 10%: resize all images to a single shared dimensions
    if max(get_all_possible_aspect_ratios(images)) < 0.1:
        resize_height, resize_width, _ = list(sorted(shapes, key=lambda x: x[1]))[len(widths)//2] # Pick shape from the image with the median width
        for i, image in enumerate(images):
            images[i] = cv2.resize(image, (resize_width, resize_height))

        # recalculate shapes
        shapes = np.stack([image.shape for image in images])
        heights, widths, channels = shapes[:, 0], shapes[:, 1], shapes[:, 2]

    assert np.all(shapes[:, 2] == 3), "All images should be RGB at this point through `parse_image_as_uint8_rgb_numpy_array(images)`"

    # ------------------------------------------
    # Find a suitable layout
    # ------------------------------------------
    height_heuristic, width_heuristic, plot_settings = None, None, "brute_force"

    # (1) All images are of identical shape
    if np.all(shapes[0, :] == shapes[1:, :]):
        plot_settings = "shape_size"
        height_heuristic, width_heuristic = shapes[0][0], shapes[0][1]

    # (2) If all images are wider than tall: resize to a shared width
    elif np.all(widths > heights):
        plot_settings = "width_shared"
        shared_width = int(np.median(widths))
        for i, image in enumerate(images):
            images[i] = image_modifier.resize_image_respect_aspect_ratio(image, shared_width, None)[0]

    # (3) If all images are taller than wide: resize to a shared height
    elif np.all(heights > widths):
        plot_settings = "height_shared"
        shared_height = int(np.median(heights))
        for i, image in enumerate(images):
            images[i] = image_modifier.resize_image_respect_aspect_ratio(image, None, shared_height)[0]

    # Calculate the final width and height heuristic.
    if (height_heuristic is None) and (width_heuristic is None):
        shapes = np.stack([image.shape for image in images])
        height_heuristic, width_heuristic = int(np.mean(shapes[:, 0])), int(np.mean(shapes[:, 1]))

    # ------------------------------------------
    # Create grid image
    # ------------------------------------------

    # Sanity check
    if (height_heuristic is None) or (width_heuristic is None):
        raise RuntimeError("This shouldn't be possible")

    max_width, max_height = 1920, 1080
    if max_output_image_size_wh is not None:
        max_width, max_height = max_output_image_size_wh

    # Get layout
    cols, rows, scaling_factor, loss_info = _get_grid_parameters(
        len(images), height=height_heuristic, width=width_heuristic, max_width=max_width, max_height=max_height, **global_config.config_auto_grid_layout
    )
    if (scaling_factor != 1.0) and allow_auto_resize:
        for i, image in enumerate(images):
            h, w, _ = image.shape
            images[i] = cv2.resize(image, (int(w * scaling_factor), int(h * scaling_factor)))

    # Show grid according to `plot_settings`
    shared_settings = {"max_output_image_size_wh":max_output_image_size_wh, "display_image":False, "c_drop_down_shadow":None,
                       "image_source":images, "rows":rows, "cols":cols, "canvas_background_color":(245,245,245), "return_image":True}
    if plot_settings == "shape_size":
        final = show_grid_configurable(**shared_settings, c_image_spacing=None, c_image_area_margins=None, c_image_resizing=None)
    elif plot_settings in ["width_shared", "height_shared"]:
        image_spacing = {"col_spacing":5, "row_spacing":5, "col_sep_line":1, "row_sep_line":1, "sep_line_color":(175, 175, 175, 0.3)}
        final = show_grid_configurable(**shared_settings, c_image_spacing=image_spacing, c_image_area_margins=None, c_image_resizing=None, c_image_border=None)
    elif plot_settings == "brute_force":
        final = show_grid_configurable(**shared_settings)
    else:
        raise RuntimeError("This shouldn't be possible")

    # ------------------------------------------
    # Wrap up
    # ------------------------------------------

    if title is not None:
        final = image_modifier.draw_image_title(final, title, top_padding_color=shared_settings["canvas_background_color"])
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, final)
    if display_image:
        show.show_image(final)
    if return_image:
        return final

__all__ = ["show_grid_configurable", "show_grid"]


from __future__ import annotations
import numpy as np
import PIL
from PIL import Image
import os
import requests
import cv2
import copy

# Mine
from config import *
import checker


def resize_and_bgr_parser(image, resize_factor, BGR2RGB):
    checker.assert_resize_factor_is_ok(resize_factor)
    if resize_factor != 1.0:
        height = int(image.shape[0] * resize_factor)
        if height < global_config.image_min_height:
            raise ValueError(f"After resize the image's `{height=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                             f"You can change this threshold through: `global_config.min_height=<NEW_VALUE>`")
        width = int(image.shape[1] * resize_factor)
        if width < global_config.image_min_width:
            raise ValueError(f"After resize the image's `{width=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                             f"You can change this threshold through: `global_config.min_width=<NEW_VALUE>`")
        interpolation = cv2.INTER_AREA if (resize_factor < 1.0) else global_config.resize_interpolation
        image = cv2.resize(image, (width, height), interpolation=interpolation)
    if BGR2RGB:
        image = image[:, :, ::-1]
    return image


# noinspection PyTypeChecker
def _parse_url(image_url:str) -> np.ndarray:
    checker.assert_type(image_url, str, "image_url")
    response = requests.get(image_url, stream=True)
    if response.status_code != 200:
        raise RuntimeError(f"`{image_url=}` was interpreted as an image URL, but the get-request failed with status code: `{response.status_code}`")
    try:
        image = np.array(Image.open(response.raw))
    except Exception as e:
        raise RuntimeError(f"Failed to convert downloaded image to an np.array with error:\n{e}")

    try:
        return _parse_numpy(image)
    except Exception as e:
        raise RuntimeError(f"`{image_url=}` was interpreted as an image URL and successfully downloaded, "
                           f"but it could not be converted to standard format (uint8, RGB np.ndarray, range:0-255) due to the error: `{e}`")


def _parse_path(image_path) -> np.ndarray:
    try:
        image = cv2.imread(image_path)[:, :, ::-1] # [:, :, ::-1] is BGR->RGB
    except Exception as e:
        raise RuntimeError(f"`{image_path=}` was interpreted as an image path, but it could not be loaded due to the error: `{e}`")

    # TODO: Is this neccesary? cv2.imread always return a correctly formatted RGB image as far as I can tell
    try:
        return _parse_numpy(image)
    except Exception as e:
        raise RuntimeError(f"`{image_path=}` was interpreted as an image path and successfully loaded, "
                           f"but it could not be converted to standard format (uint8, RGB np.ndarray, range:0-255) due to the error: `{e}`")

def parse_torch_image_batch_as_uint8_rgb_numpy_array(
        torch_batch:torch.Tensor,
        resize_factor: float = 1.0,
        BGR2RGB:bool = False,
        attempt_inverse_standardization_if_necessary:bool=True
) -> List[ndarray]:
    """
    Convert an RGB `torch_batch` (N, 3, H, W) into a list of numpy images (H, W, 3).

    :param torch_batch: The input batch of images in torch.Tensor format (N, 3, H, W). Can handle (N, H, W, 3) as well.
    :param resize_factor: Resize factor for each image. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB.
    :param attempt_inverse_standardization_if_necessary: Flag to attempt inverse standardization if necessary (ImageNet std and mean are used).
           For customization use `global_config.torch_standardization_mean` and `global_config.torch_standardization_std`.
    :return: list of numpy images (H, W, 3) in a standardized format (uint8 with pixels value between 0-255).
    """

    # Checks
    if not TORCH_FOUND:
        raise RuntimeError("You are attempting to parse a torch image, but failed to find the torch library")
    checker.assert_types(
        [torch_batch, resize_factor, BGR2RGB, attempt_inverse_standardization_if_necessary],
        [torch.Tensor, float, bool, bool],
        ["torch_batch", "resize_factor", "BGR2RGB", "attempt_inverse_standardization_if_necessary"]
    )
    checker.assert_resize_factor_is_ok(resize_factor)

    # Ensure CPU
    torch_batch = torch_batch.detach().cpu()
    if "int" in str(torch_batch.dtype):
        torch_batch = torch_batch.int()
    elif "float" in str(torch_batch.dtype):
        torch_batch = torch_batch.float()
    else:
        raise ValueError(f"Expected `{torch_batch.dtype=}` to be some variant of a float or int")

    # Dimensions
    dims = len(torch_batch.shape)
    if dims != 4:
        raise ValueError(f"Expected `torch_batch` to be of shape (batch_size, channels, height, width), but found `{torch_batch.shape=}")

    # Color channels
    batch_size, c, h, w = torch_batch.shape
    if (c >= w) or (c >= h):
        if (w < global_config.image_min_width) and (h < global_config.image_min_height) and (c == 1 or c == 3):
            msg = f"Expected `{torch_batch.shape=}` to be of format (batch_size, channel, height, width),\n" \
                  f"but found something that will be interpreted as (batch_size, height, width, channel).\n" \
                  f"`torch_batch` will be permuted to fit this assumption which will yield: `{torch_batch.permute(0,3,1,2).shape=}`"
            warnings.warn(msg)
            torch_batch = torch_batch.permute(0, 3, 1, 2)
        else:
            raise ValueError(f"Expected `torch_batch` to be of shape (batch_size, channels, height, width), but found `{torch_batch.shape=}")

    if torch_batch.shape[1] not in [1, 3]:
        raise ValueError(f"Expected `torch_batch` to be of shape (batch_size, channels, height, width), where channels in [1,3], but found `{torch_batch.shape=}")

    # Batch size
    if not (batch_size > 0):
        raise ValueError(f"Expected `torch_batch` to be of shape (batch_size, channels, height, width), where batch_size>0, but found `{torch_batch.shape=}")

    # Width and height
    if torch_batch.shape[2] < global_config.image_min_height:
        raise ValueError(f"The height of the image `{torch_batch.shape[2]=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                         f"You can change this threshold through: `global_config.min_height=<NEW_VALUE>`")
    if torch_batch.shape[3] < global_config.image_min_width:
        raise ValueError(f"The width of the image `{torch_batch.shape[3]=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                         f"You can change this threshold through: `global_config.min_width=<NEW_VALUE>`")

    # Attempt converting all images in the batch
    images = []
    for i, image in enumerate(torch_batch):
        try:
            images.append(resize_and_bgr_parser(
                _parse_torch(image,attempt_inverse_standardization_if_necessary), resize_factor, BGR2RGB)
            )
        except Exception as e:
            raise RuntimeError(f"Failed to parse the {i}/{batch_size} image in the batch due to error:\n{e}")
    return images

def parse_numpy_image_batch_as_uint8_rgb_numpy_array(numpy_batch:ndarray, resize_factor: float = 1.0, BGR2RGB:bool = False) -> List[ndarray]:
    """
    Convert an RGB `numpy_batch` (N, H, W, 3) into a list of numpy images (H, W, 3).

    :param numpy_batch: The input batch of images in torch.Tensor format (N, 3, H, W).
    :param resize_factor: Resize factor for each image. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB.
    :return: list of numpy images (H, W, 3) in a standardized format (uint8 with pixels value between 0-255).
    """

    # Checks
    checker.assert_types([numpy_batch, resize_factor, BGR2RGB], [ndarray, float, bool], ["numpy_batch", "resize_factor", "BGR2RGB"])
    if not (len(numpy_batch.shape) == 4):
        raise ValueError(f"Expected `numpy_batch` to be of shape (batch_size, height, width, 3), but found `{numpy_batch.shape=}")
    batch_size, height, width, channels = numpy_batch.shape
    if batch_size == 0:
        raise ValueError(f"Expected `numpy_batch` to be of shape (batch_size, height, width, 3) with a batch_size>0, but found `{numpy_batch.shape=}")
    if batch_size > 1_000:
        msg = f"`{batch_size=}` is suspiciously high."
        warnings.warn(msg)
    if not (channels == 3):
        raise ValueError(f"Expected `numpy_batch` to be a color image of shape (batch_size, height, width, 3), but found `{numpy_batch.shape=}")
    if height < global_config.image_min_height:
        raise ValueError(f"The height of the image `{numpy_batch.shape[1]=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                         f"You can change this threshold through: `global_config.min_height=<NEW_VALUE>`")
    if width < global_config.image_min_width:
        raise ValueError(f"The width of the image `{numpy_batch.shape[2]=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                         f"You can change this threshold through: `global_config.min_width=<NEW_VALUE>`")
    checker.assert_resize_factor_is_ok(resize_factor)

    # Attempt converting all images in the batch
    images = []
    for i, image in enumerate(numpy_batch):
        try:
            images.append(resize_and_bgr_parser(_parse_numpy(image), resize_factor, BGR2RGB))
        except Exception as e:
            raise RuntimeError(f"Failed to parse the {i}/{batch_size} image in the batch due to error:\n{e}")
    return images

def _parse_torch(torch_image:torch.Tensor, attempt_inverse_standardization_if_necessary:bool=True) -> np.ndarray:
    # ---------------------------------------------------------
    # Check and simple conversions
    # ---------------------------------------------------------
    if not TORCH_FOUND:
        raise RuntimeError("You are attempting to parse a torch image, but failed to find the torch library")
    checker.assert_type(torch_image, torch.Tensor, "torch_image")

    # Ensure CPU
    torch_image = torch_image.detach().cpu().clone()
    if "int" in str(torch_image.dtype):
        torch_image = torch_image.int()
    elif "float" in str(torch_image.dtype):
        torch_image = torch_image.float()
    else:
        raise ValueError(f"Expected `{torch_image.dtype=}` to be some variant of a float or int")

    # Dimensions
    dims = len(torch_image.shape)
    if dims not in [2,3]:
        if (dims == 4) and (torch_image.shape[0] == 1):
            msg = f"Expected `{torch_image.shape=}` to be of format (channel, height, width),\n" \
                  f"but found something that will be interpreted as (1, channel, height, width).\n" \
                  f"`torch.image` will be squeezed to fit this assumption which will yield: `{torch_image.squeeze(0).shape=}`"
            warnings.warn(msg)
            torch_image = torch_image.squeeze(0)
            dims = len(torch_image.shape)
            assert dims == 3, "Should not be possible."
        else:
            extra = f"If `torch_image` is a batch (or similar) represented as a tensor, pass each instance individually." if (dims==4) else ""
            raise ValueError(f"Expected `torch_image` to be of shape (channels, height, width) or (height, width), but found `{torch_image.shape=}`.\n{extra}")

    # Color channels
    image_height, image_width = torch_image.shape[-2], torch_image.shape[-1]
    color_channels = None if (dims == 2) else torch_image.shape[0]
    if (dims == 3) and (color_channels >= image_width or color_channels >= image_height):
        h,w,c = torch_image.shape
        if (w < global_config.image_min_width) and (h < global_config.image_min_height) and (c == 1 or c==3):
            msg = f"Expected `{torch_image.shape=}` to be of format (channel, height, width) or (height, width),\n" \
                  f"but found something that will be interpreted as (height, width, channels).\n" \
                  f"`torch.image` will be permuted to fit this assumption which will yield: `{torch_image.permute(1,2,0).shape=}`"
            warnings.warn(msg)
            torch_image = torch_image.permute(1, 2, 0)
        else:
            raise ValueError(f"Expected `torch_image` to be of shape (channels, height, width) or (height, width), but found `{torch_image.shape=}`.")

    # Width and height
    if torch_image.shape[-1] < global_config.image_min_width:
        raise ValueError(f"The width of the image `{torch_image.shape[-1]=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                         f"You can change this threshold through: `global_config.min_width=<NEW_VALUE>`")
    if torch_image.shape[-2] < global_config.image_min_height:
        raise ValueError(f"The height of the image `{torch_image.shape[-2]=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                         f"You can change this threshold through: `global_config.min_height=<NEW_VALUE>`")

    # ---------------------------------------------------------
    # Standardized
    # ---------------------------------------------------------

    # Setup
    t_min, t_max, t_std, t_mean = torch.min(torch_image), torch.max(torch_image), torch.std(torch_image.float()), torch.mean(torch_image.float())
    is_not_normalized_image = (torch_image.min() < -0.01).any() or (torch_image.max() > 1.01).any()
    range_assumed_standardized = (-10 < t_min) and (t_max < 10)
    std_and_mean_assumed_standardized = (-1 < t_std) and (t_std < 3.0) and (-2 < t_mean) and (t_mean < 2)

    assumed_standardized = is_not_normalized_image and range_assumed_standardized and std_and_mean_assumed_standardized and global_config.try_inverse_torch_standardization
    assumed_rgb_image = (len(torch_image.shape) == 3) and (torch_image.shape[0] == 3)

    # If `torch_image` is assumed to have been standardized: Attempt to inverse the standardization
    if (assumed_standardized and assumed_rgb_image) and attempt_inverse_standardization_if_necessary:
        mean = torch.as_tensor(global_config.torch_standardization_mean, dtype=torch_image.dtype).view(-1, 1, 1)
        std = torch.as_tensor(global_config.torch_standardization_std, dtype=torch_image.dtype).view(-1, 1, 1)
        torch_image.mul_(std).add_(mean)
    if (assumed_standardized and (not assumed_rgb_image)) and attempt_inverse_standardization_if_necessary:
        msg = f"`torch_image` appears to have been standardized, but cannot attempt to inverse the standardization, " \
              f"Because {torch_image.shaoe=} does not match a typical RGB tensor image of shape (3, height, width)"
        warnings.warn(msg)

    # ---------------------------------------------------------
    # Attempt np.ndarray conversion
    # ---------------------------------------------------------

    # (c, h, w) --> (h, w, c) if necessary
    if len(torch_image.shape) == 3:
        torch_image = torch_image.permute(1, 2, 0)

    try:
        return _parse_numpy(torch_image.numpy())
    except Exception as e:
        raise RuntimeError(f"Failed to convert torch-image to np.array with error:\n{e}")


def _parse_numpy(numpy_image:np.ndarray) -> np.ndarray:
    # TODO: What happens if you pass some that's (h, w, 2)?? And what should happened?
    # ---------------------------------------------------------
    # Check
    # ---------------------------------------------------------
    checker.assert_type(numpy_image, np.ndarray, "numpy_image")

    # Dimensions
    dims = len(numpy_image.shape)
    if dims not in [2,3]:
        extra = f"If `numpy_image` is a video (or similar) represented as an np.ndarray try passing each frame individually." if (dims==4) else ""
        raise ValueError(f"Expected `numpy_image` to be of shape (height, width, channels) or (height, width), but found `{numpy_image.shape=}`.\n{extra}")

    # Color channels
    image_height, image_width = numpy_image.shape[0], numpy_image.shape[1]
    color_channels = None if (dims == 2) else numpy_image.shape[2]
    if (dims == 3) and (color_channels >= image_width or color_channels >= image_height):
        raise ValueError(f"Expected `numpy_image` to be of shape (height, width, channels) or (height, width), but found `{numpy_image.shape=}`.\n"
                         f"If you have been dealing with tensors: `<YOUR_NUMPY_IMAGE>.transpose(1,2,0)` might fix the issue.")

    # Width and height
    if numpy_image.shape[1] < global_config.image_min_width:
        raise ValueError(f"The width of the image `{numpy_image.shape[1]=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                         f"You can change this threshold through: `global_config.min_width=<NEW_VALUE>`")
    if numpy_image.shape[0] < global_config.image_min_height:
        raise ValueError(f"The height of the image `{numpy_image.shape[0]=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                         f"You can change this threshold through: `global_config.min_height=<NEW_VALUE>`")

    # ---------------------------------------------------------
    # Color channels
    # ---------------------------------------------------------

    # setup
    has_alpha = (dims == 3) and numpy_image.shape[2] == 4
    is_2_channel_greyscale = dims == 2
    is_3_channel_greyscale = (dims == 3) and (numpy_image.shape[-1] == 1)

    # Remove alpha channel (for compatibility)
    if has_alpha:
        numpy_image = numpy_image[:, :, :3]

    # Adds 3 identical channels to greyscale images (for compatibility)
    elif is_3_channel_greyscale:
        numpy_image = np.repeat(numpy_image, 3, axis=2)
    elif is_2_channel_greyscale:
        numpy_image = np.repeat(np.expand_dims(numpy_image, 2), 3, axis=2)

    # ---------------------------------------------------------
    # Data type
    # ---------------------------------------------------------

    # Setup
    data_type = str(numpy_image.dtype)
    is_float = data_type.startswith("float")
    is_unsigned_int = data_type.startswith("uint")
    is_int = data_type.startswith("int")

    # If the dtype is not recognized
    if not any([is_float, is_unsigned_int, is_int]):
        msg = f"Expected `{str(numpy_image.dtype)=}` to be a uint, int or float. Will attempt auto conversion, but this may well fail!"
        warnings.warn(msg)

        # Attempt auto conversion for all 3 acceptable dtypes
        for dtype in ("uint", "int", "float"):
            try:
                numpy_image_new_type = copy.deepcopy(numpy_image).astype(dtype)
                numpy_image_new_type = _parse_numpy(numpy_image_new_type)
                return numpy_image_new_type
            except Exception as e:
                continue

        raise ValueError(f"Expected `{str(numpy_image.dtype)=}` to be a uint, int or float.")

    if is_float:
        if (0.0 <= numpy_image).all() and (numpy_image <= 1.0).all():
            numpy_image = (numpy_image * 255).astype(np.uint8)
        elif (0.0 <= numpy_image).all() and (numpy_image <= 255.0).all():
            numpy_image = numpy_image.astype(np.uint8)
        elif (global_config.image_normalization_clip_min <= numpy_image).all() and (numpy_image <= global_config.image_normalization_clip_max).all(): # TODO: Check
            msg = f"`numpy_image` is assumed to be a normalized (pixel values within 0-1) image, but one or more pixel values had to be clipped because they were outside 0-1." \
                  f"You can modify the minimum and maximum thresholds to control when clipping is allowed through: " \
                  f"`global_config.image_normalization_clip_min=<NEW_VALUE>` and `global_config.image_normalization_clip_max=<NEW_VALUE>`"
            warnings.warn(msg)
            numpy_image = (numpy_image.clip(0,1) * 255).astype(np.uint8)

    elif is_unsigned_int or is_int:
        # If black and white / binary image, cast to 0-255 range
        if np.all(np.logical_or(numpy_image == 0, numpy_image == 1)):
            numpy_image = (numpy_image * 255).astype(np.uint8)
        elif (0 <= numpy_image).all() and (numpy_image <= 255).all():
            numpy_image = numpy_image.astype(np.uint8)
        elif (global_config.image_clip_min <= numpy_image).all() and (numpy_image <= global_config.image_clip_max).all(): # TODO: Check
            msg = f"`numpy_image` is assumed to be a normal (pixel values within 0-255) image, but one or more pixel values had to be clipped because they were outside 0-255." \
                  f"You can modify the minimum and maximum thresholds to control when clipping is allowed through: " \
                  f"`global_config.image_clip_min=<NEW_VALUE>` and `global_config.image_clip_max=<NEW_VALUE>`"
            warnings.warn(msg)
            numpy_image = (numpy_image.clip(0, 255)).astype(np.uint8)
        # TODO: add checks for High Dynamic Range (HDR) 16/24 bits pictures

    return numpy_image


# noinspection PyTypeChecker
def _parse_pillow(image_pillow:PIL.Image.Image) -> np.ndarray:
    checker.assert_type(image_pillow, PIL.Image.Image, "image_pillow")
    try:
        image_pillow = np.array(image_pillow)
    except Exception as e:
        raise RuntimeError(f"Failed to convert PIL-image to np.array with error:\n{e}")
    return _parse_numpy(image_pillow)

def parse_image_as_uint8_rgb_numpy_array(image_source: ImageType, resize_factor: float = 1.0, BGR2RGB:bool = False) -> ndarray:
    """
    Convert `image_source` into a standardized numpy array - shape: (H, W, 3), pixels: uint8 with pixels value between 0-255.

    :param image_source: An image. Can handle most image sources: ndarray, url, path, torch, PIL. For a full description see: TODO insert link
    :param resize_factor: Resize factor for each image. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB.
    :return: a numpy image (H, W, 3) in a standardized format (uint8 with pixels value between 0-255).
    """

    # Type checks
    checker.assert_image_type(image_source, "image_source")
    checker.assert_types([resize_factor, BGR2RGB], [float, bool], ["resize_factor", "BGR2RGB"])
    checker.assert_resize_factor_is_ok(resize_factor)

    # Determine what kind of data we're dealing with
    is_path = os.path.exists(image_source) if isinstance(image_source, str) else False
    is_url = True if (isinstance(image_source, str) and checker.is_valid_url(image_source)) else False
    is_ndarray = True if isinstance(image_source, np.ndarray) else False
    is_torch_tensor = True if (TORCH_FOUND and isinstance(image_source, torch.Tensor)) else False
    is_pillow = True if isinstance(image_source, PIL.Image.Image) else False

    # Sanity checks
    if is_path and is_url:
        raise AssertionError("This should not be possible. Don't see how `image_source` can be a path and a url simultaneously.")
    if not any([is_path, is_url, is_ndarray, is_torch_tensor, is_pillow]):
        if isinstance(image_source, str) and (not os.path.exists(image_source)):
            raise ValueError(f"`{image_source=}` is a string, but failed to detect it as a valid path/URL. Did you perhaps provide a bad path/url?")
        raise AssertionError(f"This should not be possible. `image_source` of type `{type(image_source)=}` should have been accepted/rejected as a valid/invalid type by now.")

    # Distribute according to detected type
    if is_path:
        image = _parse_path(image_source)
    elif is_url:
        image = _parse_url(image_source)
    elif is_ndarray:
        image = _parse_numpy(image_source)
    elif is_torch_tensor:
        image = _parse_torch(image_source)
    elif is_pillow:
        image = _parse_pillow(image_source)
    else:
        raise AssertionError("This should not be possible.")

    # Remove potential alpha channels
    if image.shape[2] == 4:
        image = image[:, :, :3]

    shape_info = image_source.shape if hasattr(image_source, "shape") else None
    extra_debug_info = \
         f"\nHere's some debug info for the image: " \
         f"\n{'-'*25 + ' types ' + '-'*25}\n"\
         f"type (input)  =\t{type(image_source)}\n"\
         f"type (output) =\t{type(image.dtype)}" \
         f"\n\n{'-' * 25 + ' shape ' + '-' * 25}\n" \
         f"shape (input)  =\t{shape_info}\n" \
         f"shape (output)  =\t{image.shape}" \
         f"\n\n{'-' * 25 + ' pixel ' + '-' * 25}\n" \
         f"min=\t{np.min(image)}\n"\
         f"max=\t{np.max(image)}\n"\
         f"mean=\t{np.mean(image)}\n" \
         f"median=\t{np.median(image)}\n"\
         f"std=\t{np.std(image)}\n"


    # Final check
    if not isinstance(image, np.ndarray):
        raise ValueError(f"Failed to convert `{type(image_source)=}` to an np.ndarray for unknown reasons.{extra_debug_info}")
    if (image < 0).any() or (image > 255).any():
        raise ValueError(f"The pixel values of `image_source` are not valid and all automatic conversions failed.{extra_debug_info}")
    if image.dtype != np.uint8:
        raise ValueError(f"Failed to convert `{image_source.dtype=}` to dtype uint8 for unknown reasons.{extra_debug_info}")
    if image.shape[1] < global_config.image_min_width:
        raise ValueError(f"The width of the image `{image.shape[1]=}` is below the minimum threshold `{global_config.image_min_width}`.\n"
                         f"You can change this threshold through: `global_config.image_min_width=<NEW_VALUE>`")
    if image.shape[0] < global_config.image_min_height:
        raise ValueError(f"The height of the image `{image.shape[0]=}` is below the minimum threshold`{global_config.image_min_height}`.\n"
                         f"You can change this threshold through: `global_config.image_min_height=<NEW_VALUE>`")
    if (len(image.shape) != 3) or (image.shape[2] != 3):
        raise ValueError(f"Failed to convert `image_source` to standard shape (height, width, 3): `{image.shape=}`.{extra_debug_info}")

    # Modifications
    if resize_factor != 1.0:
        width = int(image.shape[1] * resize_factor)
        height = int(image.shape[0] * resize_factor)
        interpolation = cv2.INTER_AREA if (resize_factor < 1.0) else global_config.resize_interpolation
        image = cv2.resize(image, (width, height), interpolation=interpolation)
    if BGR2RGB:
        image = image[:, :, ::-1]

    return image


def threshold_image_count(image_source):
    warning_msg = f"Received `{len(image_source)}` images, the maximum limit is `{global_config.show_max_image_amount}`. " \
      f"Will pick `{global_config.show_max_image_amount}` random images from `image_source` for display and discard the rest.\n" \
      f"If you wish to change this limit, you can set `global_config.show_max_image_amount = <NEW LIMIT>`"
    warnings.warn(warning_msg)
    random_indexes = np.random.choice(np.arange(len(image_source)), global_config.show_max_image_amount, replace=False)
    images = [image_source[i] for i in sorted(random_indexes)]
    return images

def parse_arbitrary_image_source(image_source:ImageSource, resize_factor: float = 1., BGR2RGB: bool = False) -> List[ndarray]:
    """
    Convert image(s) from `image_source` into a standardized numpy format - shape: (H, W, 3), pixels: uint8 with pixels value between 0-255.

    :param image_source: A image or list of images. Can handle most image sources: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param resize_factor: Resize factor for each image. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB.
    :return: a list of numpy images (H, W, 3) in a standardized format (uint8 with pixels value between 0-255).
    """

    # Checks
    if not torch:
        checker.assert_in(type(image_source), [np.ndarray, PIL.Image.Image, str, list, tuple], "image_source")
    else:
        checker.assert_in(type(image_source), [np.ndarray, torch.Tensor, PIL.Image.Image, str, list, tuple], "image_source")
    checker.assert_types([resize_factor, BGR2RGB], [float, bool], ["resize_factor", "BGR2RGB"])
    if isinstance(image_source, (list, tuple)):
        if len(image_source) == 0:
            raise ValueError("`image_source` should contain image-information, but it's empty. Did you perhaps pass in an empty list or something similar?")
        for i, item in enumerate(image_source):
            checker.assert_image_type(item, f"image_source[{i}]")
    checker.assert_resize_factor_is_ok(resize_factor)


    # Prepare `image_source` for parsing
    if not isinstance(image_source, (list, tuple)):
        image_source = [image_source]
    if len(image_source) > global_config.show_max_image_amount:
        image_source = threshold_image_count(image_source)

    # Parse `image_source`
    images = []
    for item in image_source:
        if (not isinstance(item, PIL.Image.Image)) and ((item is None) or (len(item) == 0)):
            raise ValueError(f"`image_source` itself or one of its items are invalid: `{item=}`")
        elif isinstance(item, torch.Tensor) and checker.assumed_torch_batch(item):
            images += parse_torch_image_batch_as_uint8_rgb_numpy_array(item, resize_factor, BGR2RGB)
        elif isinstance(item, np.ndarray) and checker.assumed_numpy_batch(item):
            try:
                images += parse_numpy_image_batch_as_uint8_rgb_numpy_array(item, resize_factor, BGR2RGB)
            except Exception as e:
                raise ValueError(f"Received a numpy array of shape {item.shape=} which was interpreted as a batch of images (N, height, width, channel)."
                                 f"But failed to parse it as such. Error:\n{e}")
        elif isinstance(item, str) and any(item.lower().endswith(f) for f in global_config.video_formats_all):
            import video
            images.append(
                video.show_video(
                    item, verbose=global_config.in_debug_mode, display_image=False, return_image=True,
                    max_output_image_size_wh=None, resize_factor=resize_factor, BGR2RGB=BGR2RGB
                )
            )
        else:
            images.append(parse_image_as_uint8_rgb_numpy_array(item, resize_factor, BGR2RGB))
        if len(images) > global_config.show_max_image_amount:
            images = threshold_image_count(images)
            break

    assert len(images) > 0, "This should not be possible"
    return images

def quick_parse_image_source(image_source:ImageSource) -> list[ndarray]:
    if not isinstance(image_source, (tuple, list)):
        image_source = [image_source]
    if len(image_source) > global_config.show_max_image_amount:
        image_source = threshold_image_count(image_source)
    for item in image_source:
        checker.quick_assert_valid_image(item)
    return image_source


__all__ = [
    "parse_torch_image_batch_as_uint8_rgb_numpy_array",
    "parse_numpy_image_batch_as_uint8_rgb_numpy_array",
    "parse_image_as_uint8_rgb_numpy_array",
    "parse_arbitrary_image_source"
]


from __future__ import annotations
from PIL import Image
import cv2

from config import *
import utils
import checker
import image_modifier
import grid
import collage
import parsers


def show_image(
    image_source:ImageType,
    title:Optional[str]=None,
    display_image: bool = True,
    return_image: bool = False,
    max_output_image_size_wh: tuple = None,
    save_image_path:Optional[str] = None,
    resize_factor: float = 1.0,
    BGR2RGB: bool = False,
    parse_image:bool=True
) -> Optional[ndarray]:
    """
    Display a single image from `image_source`, regardless of its format.

    :param image_source: An image. Can handle most image sources: ndarray, url, path, torch, PIL. For a full description see: TODO insert link
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.
    :param parse_image: If True, will thoroughly check image_source and ensure it's in the correct format

    """

    # Init
    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image)
    if parse_image:
        image_source = parsers.parse_image_as_uint8_rgb_numpy_array(image_source, resize_factor, BGR2RGB)
    else:
        checker.quick_assert_valid_image(image_source)

    # Universal stuff
    if title is not None:
        image_source = image_modifier.draw_image_title(image_source, title)
    if max_output_image_size_wh is not None:
        image_source = image_modifier.resize_universal_output_image(image_source, max_output_image_size_wh=max_output_image_size_wh)
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, image_source)

    if display_image:
        # Display in jupyter notebook
        if IN_JUPYTER:
            try:
                display(Image.fromarray(image_source))
            except Exception as e:
                warning_message = f"Failed to display `image` in jupyter due to the following error: `{e}`."
                warnings.warn(warning_message)

        # Display with cv2
        else:
            try:
                cv2.imshow("Just show it", image_source[:, :, ::-1]) # [:, :, ::-1] converts RGB -> BGR
                cv2.waitKey(global_config.cv2_wait_key_delay_ms)
                cv2.destroyAllWindows()
            except Exception as e:
                warning_message = f"Failed to display `image` with cv2 due to the following error: `{e}`."
                warnings.warn(warning_message)
                cv2.destroyAllWindows()

    if return_image:
        return image_source


def show(
    image_source:ImageSource,
    title: Optional[str] = None,
    display_image: bool = True,
    return_image: bool = False,
    max_output_image_size_wh: Optional[tuple] = (1920, 1080),
    save_image_path: Optional[str] = None,
    resize_factor: float = 1.0,
    BGR2RGB: bool = False,
    parse_image: bool = True,
):

    """
    Display `image_source` in an appropriate way, regardless of its format.

    :param image_source: An image or list of images. Can handle most image sources: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.
    :param parse_image: If True, will thoroughly check image_source and ensure it's in the correct format
    :returns: If return_image is True, return an image of shape np.ndarray (H, W, 3), otherwise None
    """

    # Checks
    # --------------------------------------------------------------------
    # # Universal stuff
    # --------------------------------------------------------------------

    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image)
    if parse_image:
        image_source = parsers.parse_arbitrary_image_source(image_source, resize_factor, BGR2RGB)
    else:
        image_source = parsers.quick_parse_image_source(image_source)

    # Find out how best to display the images extracted from `source`
    if isinstance(image_source, list) and (len(image_source) == 1):
        final_image = image_source[0]
    elif max(grid.get_all_possible_aspect_ratios(image_source)) < 0.1:
        final_image = grid.show_grid(image_source, display_image=False, return_image=True, max_output_image_size_wh=(3000, 3000))
    else:
        final_image = collage.show_collage(image_source, display_image=False, return_image=True, max_output_image_size_wh=(3000, 3000))

    # Wrap up
    if title:
        final_image = image_modifier.draw_image_title(final_image, title)
    if max_output_image_size_wh is not None:
        final_image = image_modifier.resize_universal_output_image(final_image, max_output_image_size_wh=max_output_image_size_wh)
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, final_image)
    if display_image:
        show_image(final_image)
    if return_image:
        return final_image


__all__ = ["show_image", "show"]


from __future__ import annotations
import os
from PIL import ImageFont
import cv2
import numpy as np
from datetime import datetime
from os.path import getmtime

from config import *
import checker


def save_image_to_disk(path:str, image:ndarray) -> None:
    checker.assert_types([path, image], [str, ndarray], ["path", "image"])
    if not any(path.lower().endswith(f) for f in global_config.image_formats):
        raise ValueError(f"`{path=}` must be suffixed by: `{global_config.image_formats}`, but found neither.")
    cv2.imwrite(path, cv2.cvtColor(image, cv2.COLOR_RGB2BGR))


def get_font(font_size:int, font_thickness:int, italic:bool=False):
    checker.assert_types([font_size, font_thickness, italic], [int, int, bool], ["font_size", "font_thickness", "italic"])
    checker.assert_positive_int(font_size, variable_name="font_size")
    if not (0 < font_thickness <= 7):
        raise ValueError(f"`{font_thickness=}` is not valid. As of writing this, `font_thickness` must be in [1,2,3,4,5,6,7]")

    # TODO: add return types
    # Maps from (font_thickness, is_italic) to font
    font_name_mapper = {
        (1, False): 'JetBrainsMono-Thin.ttf',
        (1, True):  'JetBrainsMono-ThinItalic.ttf',
        (2, False): 'JetBrainsMono-ExtraLight.ttf',
        (2, True):  'JetBrainsMono-ExtraLightItalic.ttf',
        (3, False): 'JetBrainsMono-Light.ttf',
        (3, True):  'JetBrainsMono-LightItalic.ttf',
        (4, False): 'JetBrainsMono-Medium.ttf',
        (4, True):  'JetBrainsMono-MediumItalic.ttf',
        (5, False): 'JetBrainsMono-SemiBold.ttf',
        (5, True):  'JetBrainsMono-SemiBoldItalic.ttf',
        (6, False): 'JetBrainsMono-Bold.ttf',
        (6, True):  'JetBrainsMono-BoldItalic.ttf',
        (7, False): 'JetBrainsMono-ExtraBold.ttf',
        (7, True):  'JetBrainsMono-ExtraBoldItalic.ttf'
    }

    # Load font from path
    font_name = font_name_mapper[(font_thickness, italic)]
    current_folder_path = os.path.dirname(__file__)
    absolute_font_path = os.path.join(current_folder_path, "__fonts", font_name) # This is probably overkill, but I don't want to deal with any relative path shenanigans
    if os.path.exists(absolute_font_path):
        try:
            font = ImageFont.truetype(absolute_font_path, size=font_size)
            return font
        except Exception as e:
            msg = f"Failed to load `{font_name=}` from path: `{absolute_font_path}`.\n" \
            f"This should not be possible because all __fonts are validated before reaching this point. " \
            f"So, I'm uncertain why the font cannot be located..\n" \
            f"The error:\n`{e}`"
            raise RuntimeError(msg)
    # TODO: Should there be a third option here that tries to download the font directly from the github page?
    # TODO: Find out if it makes better sense to try the encoding first and then the path
    # Load font directly from encodings
    else:
        from encoded_font_data import fonts_encoded
        import base64
        import io
        try:
            encoded_font_data = fonts_encoded[font_name]
            binary = base64.b64decode(encoded_font_data)  # decode from base64 to binary
            FileLike = io.BytesIO(binary)  # wrap in BytesIO to make file-like object
            font = ImageFont.truetype(FileLike, size=font_size)  # load font
            return font
        except Exception as e:
            msg = f"Failed to load `{font_name=}` from path: `{absolute_font_path}` and from pre-made encodings.\n" \
                  f"This should not be possible because all __fonts are validated before reaching this point. " \
                  f"So, I'm uncertain why the font cannot be located..\n" \
                  f"The error:\n`{e}`"
            raise RuntimeError(msg)


def frame_can_be_read(cap, index):
    cap.set(cv2.CAP_PROP_POS_FRAMES, index)
    successful, _ = cap.read()
    if successful:
        return True
    return False


def is_valid_end_frame(cap, index):
    return frame_can_be_read(cap, index) and (not frame_can_be_read(cap, index+1))


def get_cv2_video_capture(video_path: str, max_iteration: int = 75) -> Tuple[cv2.VideoCapture, dict]:
    """
    "Ensure successful loading of cv2.VideoCapture(video_path) and provide some video information, including a reliable frame count."

    :param video_path: Path to a video file
    :param max_iteration: If the frame count provided by cv2 is unreliable, binary search will be utilized to determine the frame count with a maximum number of iterations specified by max_iteration.
    :return: Successfully loaded cv2.VideoCapture and a dict with video information
    """

    checker.assert_path(video_path, "video_path")
    checker.assert_positive_int(max_iteration, variable_name="max_iteration")

    # -----------------------------------
    # Setup
    # -----------------------------------

    # Load and check video stream
    try:
        cap = cv2.VideoCapture(video_path)
        successful, _ = cap.read()
        if not successful:
            raise RuntimeError("Failed to load the first frame from `cv2.VideoCapture(video_path)`")
    except Exception as e:
        raise RuntimeError(f"Failed to load the first frame from `cv2.VideoCapture(video_path)` with error: `{e}`")

    # Video info
    info = {}
    info["frame_count"] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    info["video_width"] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    info["video_height"] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    info["fps"] = round(cap.get(cv2.CAP_PROP_FPS), 2)
    info["duration_sec"] = info["frame_count"] / info["fps"]
    info["save_time"] = datetime.fromtimestamp(getmtime(video_path)).strftime('%d/%m/%Y %H:%M:%S')
    info["cv2_is_unstable"] = False
    info["as_string"] = f'saved: {info["save_time"]}  ' \
                        f'duration: {info["duration_sec"]} sec  ' \
                        f'frames: {info["frame_count"]}  ' \
                        f'WH: {(info["video_width"], info["video_height"])}  ' \
                        f'fps: {info["fps"]}'

    # ---------------------------------------
    # Simple attempt to determine frame count
    # ---------------------------------------

    # If cv2's own frame count is correct there's no need to do anything fancy
    if is_valid_end_frame(cap, info["frame_count"]):
        cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
        return cap, info
    else:
        # msg = f"cv2's frame count is unreliable, this may be an indication that cv2 is struggling to read `{video_path=}`."
        # warnings.warn(msg)
        info["cv2_is_unstable"] = True

    # ---------------------------------------
    # Binary search to determine frame count
    # ---------------------------------------

    # Find some reasonable starting values
    start_value = 0
    for end_value in [1, 100, 1000, 5000, 10_000, 100_000, int(1e6), int(1e7), int(1e8), int(1e9), int(1e10)]:
        failed_to_read = not frame_can_be_read(cap, end_value)
        if failed_to_read:
            break
        start_value = end_value

    # Binary search
    counter = 0
    while True and (counter <= max_iteration):
        counter += 1
        seperation_point = start_value + ((end_value - start_value) // 2)
        if not frame_can_be_read(cap, seperation_point):
            end_value = seperation_point
        else:
            start_value = seperation_point
        if (end_value - start_value) == 1:
            assert is_valid_end_frame(cap, start_value), "Something went wrong"
            info["frame_count"] = end_value
            info["duration_sec"] = round(info["frame_count"] / info["fps"], 2)
            info["as_string"] = f'saved: {info["save_time"]}  ' \
                                f'duration: {info["duration_sec"]} sec  ' \
                                f'frames: {info["frame_count"]}  ' \
                                f'WH: {(info["video_width"], info["video_height"])}  ' \
                                f'fps: {info["fps"]}'
            break
    if (counter > max_iteration) and (not is_valid_end_frame(cap, start_value)):
        raise RuntimeError(f"Failed to find the total number of frames after `{max_iteration}` iterations")

    cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
    return cap, info


def extract_equally_spaced_numbers(indexes, n):
    if n < 2:
        raise ValueError(f"`{n=}` should be greater than or equal to 2.")
    if len(indexes) < n or len(indexes) < 2:
        raise ValueError(f"Expected: `{n=}` <= `{len(indexes)=}` > 2")

    first_frame = indexes[0]
    last_frame = indexes[-1]

    # We subtract 2 from n to account for the inclusion of the first and last elements.
    indices = np.linspace(0, len(indexes) - 1, num=n, endpoint=True)
    extracted_indexes = [indexes[int(i)] for i in indices]

    # Ensure that the first and last frames are included.
    extracted_indexes[0] = first_frame
    extracted_indexes[-1] = last_frame
    assert len(extracted_indexes) == n, "Should not be possible"
    return extracted_indexes

__all__ = ["get_cv2_video_capture"]


from __future__ import annotations
import numpy as np

from config import *
import show
import utils
import checker
import parsers
import image_modifier
RectType = List[int]

def _calculate_overlap(rect1:RectType, rect2:RectType) -> Tuple[int, float, float]:
    """
    >> calculate_overlap([0, 50, 100, 150], [50, 0, 150, 100])
    """
    x1, y1, x2, y2 = rect1
    x3, y3, x4, y4 = rect2

    if (x1 >= x4) or (x2 <= x3) or (y1 >= y4) or (y2 <= y3):
        return 0, 0.0, 0.0  # No overlap

    # Calculate the overlapping region
    x_left = max(x1, x3)
    y_top = max(y1, y3)
    x_right = min(x2, x4)
    y_bottom = min(y2, y4)

    # Calculate the number of overlapping pixels for each rectangle
    overlap_rect1 = (x_right - x_left) * (y_bottom - y_top)
    overlap_rect2 = overlap_rect1

    # Calculate the percentage of overlap for each rectangle
    area_rect1 = (x2 - x1) * (y2 - y1)
    overlap_percentage_rect1 = overlap_rect1 / area_rect1

    area_rect2 = (x4 - x3) * (y4 - y3)
    overlap_percentage_rect2 = overlap_rect2 / area_rect2

    assert overlap_rect1 == overlap_rect2, "The intersection should be the same regardless of rectangle"
    return overlap_rect1, overlap_percentage_rect1, overlap_percentage_rect2


def _simple_intersect_check(rect1:RectType, rect2:RectType) -> bool:
    x1, y1, x2, y2 = rect1
    x3, y3, x4, y4 = rect2
    if (x1 >= x4) or (x2 <= x3) or (y1 >= y4) or (y2 <= y3):
        return False  # No overlap
    return True  # Overlap


def _intersects(rect1:RectType, rect2:RectType, overlap_tolerance_percentage_01:float=0.0, overlap_tolerance_pixels:int=np.inf) -> bool:
    # Simple check (cheaper)
    if (overlap_tolerance_percentage_01 == 0.0) and (overlap_tolerance_pixels == np.inf):
        return _simple_intersect_check(rect1, rect2)

    # Detailed check
    intersection_pixels, overlap_rect1_percentage, overlap_rect2_percentage = _calculate_overlap(rect1, rect2)
    if intersection_pixels > overlap_tolerance_pixels:
        return True
    elif max(overlap_rect1_percentage, overlap_rect2_percentage) > overlap_tolerance_percentage_01:
        return True
    return False


def _get_packed(rectangles:Tuple[Tuple[int, int]], canvas_width:int, canvas_height:int, overlap_tolerance_percentage_01:float, collage_sort:str, add_extra_corner:bool):
    """
    # TODO: Write docstring
    # TODO Update drawio: `add_extra_corner` can add an extra corner
    :return:

    Algorithm is explained in `../doc/packing_algo.png`
    """

    # -------------------------
    # Checks
    # -------------------------
    checker.assert_types([rectangles, canvas_width, canvas_height], [tuple, int, int], ["rectangles", "canvas_width", "canvas_height"])
    if len(rectangles) < 1:
        raise ValueError(f"Expected to find at least one rectangle, but received `{len(rectangles)=}`")
    if len(rectangles) > 1_000:
        warning_message = f"Received a suspiciously large number of rectangles: `{len(rectangles)=}`"
        warnings.warn(warning_message)
    extra_rectangles_message = "`rectangles` is expected to contain tuples of two positive integers"
    if any([(len(r) != 2) for r in rectangles]):
        raise ValueError(f"One or more entries in `rectangles` has a length different then 2. \n{extra_rectangles_message}")
    if any([( (type(w)!=type(h)!=int) or w < 1 or h < 1 ) for (w, h) in rectangles]):
        raise ValueError(f"One or more widths/heights in `canvas_domains` contains an unexpected value. \n{extra_rectangles_message}")

    # -------------------------
    # Packing
    # -------------------------

    # Setup
    rectangles = [(w, h, i) for i, (w, h) in enumerate(rectangles)]
    solutions = []
    if collage_sort == "as_is":
        pass
    elif collage_sort == "area_width_height":
        rectangles = sorted(rectangles, key=lambda r: (r[0]*r[1], r[0], r[1]), reverse=True) # Sort by area, then width, then height
    elif collage_sort == "area_height_width":
        rectangles = sorted(rectangles, key=lambda r: (r[0]*r[1], r[1], r[0]), reverse=True)
    elif collage_sort == "width_height_area":
        rectangles = sorted(rectangles, key=lambda r: (r[0], r[1], r[0]*r[1]), reverse=True)
    elif collage_sort == "height_width_area":
        rectangles = sorted(rectangles, key=lambda r: (r[1], r[0], r[0]*r[1]), reverse=True)
    else:
        raise RuntimeError("Shouldn't have gotten this far")

    corners = [(0, 0)]
    fitness_mask = np.zeros((canvas_height, canvas_width), dtype=np.uint8)

    # Attempt to pack `images` into a canvas of shape (canvas_width, canvas_height)
    for (rect_width, rect_height, rect_id) in rectangles:
        for i, (x_corner, y_corner) in enumerate(corners):
            x2 = x_corner + rect_width
            y2 = y_corner + rect_height
            this_box = [x_corner, y_corner, x2, y2]
            no_intersection = not any([_intersects(this_box, other_box[:4], overlap_tolerance_percentage_01) for other_box in solutions])
            if (0 < x2 < canvas_width) and (0 < y2 < canvas_height) and no_intersection:
                corners.pop(i)
                solutions.append([x_corner, y_corner, x2, y2, rect_id])
                corners.append((x2, y_corner))  # Upper right corner
                corners.append((x_corner, y2))  # Lower left corner
                if add_extra_corner and (y_corner > 0) and (fitness_mask[y_corner-1, x2] != 1): # Not on image border / other rectangles border
                    strip = fitness_mask.copy()[:y_corner, (x2-1)] # Imagine you send a ray from the upper left corner of the just place rectangle. This finds the y of the first intersection.
                    hit = strip[::-1].argmax()
                    y2_new = (y_corner - hit) if (hit != 0) else 0
                    corners.append((x2, y2_new))  # Corner right above "Upper right corner"

                fitness_mask[y_corner:y2, x_corner:x2] = 1
                break
        corners = sorted(corners, key=lambda c: (c[0], c[1]))  # Sort corners by width then by height # TODO add a "if can potentially fill a hole" --> First

    if len(solutions) != len(rectangles):
        return None

    # Fitness
    solutions = np.array(solutions)
    max_x, max_y = solutions[:, 2:4].max(axis=0)
    fitness = fitness_mask[:max_y, :max_x].mean()

    assert len(np.unique(solutions[:, -1])) == solutions.shape[0] == len(rectangles)
    return {"fitness":fitness, "max_x":max_x, "max_y":max_y, "solutions":solutions}


def show_collage(
        image_source: ImageSource,
        canvas_domains:Tuple[Tuple[int, int]]=None,
        crop_smallest_fit:bool=True,
        overlap_tolerance:float=0.1,
        title: Optional[str] = None,
        display_image: bool = True,
        return_image: bool = False,
        max_output_image_size_wh: tuple = None,
        save_image_path: Optional[str] = None,
        resize_factor: float = 1.0,
        BGR2RGB: bool = False,
        parse_image: bool = True
) -> Optional[ndarray]:
    """
    "Pack and display all images from `image_source` into the smallest canvas available from `canvas_domains`."

    :param image_source: An image or list of images. Can handle most image sources: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param canvas_domains: A tuple of (width, height)-pairs serving as canvases to fit the images. The smallest canvas that can contain all images will be used.
                           If None: canvas_domains = tuple((int(1920 * 10 / 1.25 ** i), int(1080 * 10 / 1.25 ** i)) for i in range(16, 0, -1))
    :param crop_smallest_fit: If True, will crop the final image to avoid unused space. This means the shape of the final image can differ from those specified in canvas_domains
    :param overlap_tolerance: The maximum overlap two images may have - percentage 0-1. Note if one image overlaps 40% while the other only 10%, 40% will be considered.
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.
    :param parse_image: If True, will thoroughly check image_source and ensure it's in the correct format

    :returns: If return_image is True, return an image of shape np.ndarray (H, W, 3), otherwise None
    """

    # --------------------------------------------------------------------
    # Universal stuff
    # --------------------------------------------------------------------

    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, parse_image)
    if parse_image:
        image_source = parsers.parse_arbitrary_image_source(image_source, resize_factor, BGR2RGB)
    else:
        image_source = parsers.quick_parse_image_source(image_source)

    # Function specific
    checker.assert_types(
        [canvas_domains, crop_smallest_fit, overlap_tolerance],
        [tuple, bool, float],
        ["canvas_domains", "crop_smallest_fit", "overlap_tolerance"],
        [1,0,0]
    )

    # `canvas_domains`
    if canvas_domains is None:
        canvas_domains = tuple((int(1920 * 10 / 1.25 ** i), int(1080 * 10 / 1.25 ** i)) for i in range(16, 0, -1))
    if len(canvas_domains) < 1:
        raise ValueError(f"Expected to find at least one canvas domain, but received `{len(canvas_domains)=}`")
    if len(canvas_domains) > 1_000:
        warning_message = f"Received a suspiciously large number of canvas domains: `{len(canvas_domains)=}`"
        warnings.warn(warning_message)
    extra_canvas_message = "`canvas_domains` is expected to contain tuples of two positive integers"
    if any([(len(cd) != 2) for cd in canvas_domains]):
        raise ValueError(f"One or more entries in `canvas_domains` has a length different then 2. \n{extra_canvas_message}")
    if any([ ( (type(w)!=type(h)!=int) or w < 1 or h < 1 ) for (w, h) in canvas_domains]):
        raise ValueError(f"One or more widths/heights in `canvas_domains` contains an unexpected value. \n{extra_canvas_message}")

    max_canvas_width = max([w for (w, _) in canvas_domains])
    max_image_width = max([image.shape[1] for image in image_source])
    if max_canvas_width < max_image_width:
        raise RuntimeError(f"`{max_canvas_width=} < {max_image_width=}`")

    max_canvas_height = max([h for (_, h) in canvas_domains])
    max_image_height = max([image.shape[0] for image in image_source])
    if max_canvas_height < max_image_height:
        raise RuntimeError(f"`{max_canvas_height=} < {max_image_height=}`")

    # `overlap_tolerance`
    checker.assert_positive_float(overlap_tolerance, zero_allowed=True, max_value_allowed=0.99, variable_name="overlap_tolerance")

    # -------------------------
    # Packing
    # -------------------------

    # Setup
    rectangles = tuple((int(s.shape[1]), int(s.shape[0])) for s in image_source)
    canvas_image = None
    best_fitness, best_packing_solution = -1.0, None

    # Attempt to pack all images into the smallest of the predefined width-height combinations
    for (canvas_width, canvas_height) in canvas_domains:
        if (canvas_width < max_image_width) or (canvas_height < max_image_height): # Can safely be skipped because it will never fit
            continue

        for sorting_algorithm in ["as_is", "area_width_height", "area_height_width", "width_height_area", "height_width_area"]: # ["as_is"]:
            for add_extra_corner in [False, True]:
                results = _get_packed(rectangles, canvas_width, canvas_height, overlap_tolerance, sorting_algorithm, add_extra_corner)

                if (results is not None) and (results["fitness"] > best_fitness): # No solution found
                    best_fitness = results["fitness"]
                    best_packing_solution = results
                    best_packing_solution["sorting_algorithm"] = sorting_algorithm
                    best_packing_solution["add_extra_corner"] = add_extra_corner
                    best_packing_solution["canvas_shape"] = (canvas_width, canvas_height)

        # No solution was found, try with a bigger canvas
        if best_packing_solution is None:
            continue

        # Create a grey canvas according to the specified dimensions
        canvas_image = np.zeros((canvas_height, canvas_width, 3), dtype=np.uint8) + 65  # 65 seems to be a pretty versatile grey color i.e. it looks decent no matter the pictures

        if global_config.in_debug_mode:
            [print(f"{k}: {v}") for k,v in best_packing_solution.items() if (k != "solutions")]

        # (top_left_x, top_left_y, bottom_right_x, bottom_right_y)
        for (x1, y1, x2, y2, rid) in best_packing_solution["solutions"]:
            image = image_source[rid]
            assert image.shape[:-1] == (y2 - y1, x2 - x1)
            canvas_image[y1:y2, x1:x2, :] = image
        break


    # -------------------------
    # Wrapping up
    # -------------------------
    if canvas_image is None:
        raise RuntimeError("Failed to produce a mosaic image. There can be several reasons for this:\n"
                           "(1) Too large images\n"
                           "(2) Too many images\n"
                           "(3) The widths and heights in `canvas_domains` may be poorly chosen")
    if (best_packing_solution["max_x"] == -1) or (best_packing_solution["max_y"] == -1):
        raise RuntimeError("This should not be possible.")

    if crop_smallest_fit:
        canvas_image = canvas_image[:best_packing_solution["max_y"], :best_packing_solution["max_x"], :]

    # Wrap up
    if title:
        canvas_image = image_modifier.draw_image_title(canvas_image, text=title)
    if max_output_image_size_wh is not None:
        canvas_image = image_modifier.resize_universal_output_image(canvas_image, max_output_image_size_wh=max_output_image_size_wh)
    if display_image:
        show.show_image(canvas_image)
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, canvas_image)
    if return_image:
        return canvas_image

__all__ = ["show_collage"]


from __future__ import annotations

import numpy as np
import cv2
import os

# Mine
from config import *
import checker
import image_modifier
import parsers
import utils
import video_player
import grid
import show

image_formats = global_config.image_formats
video_formats_all = global_config.video_formats_all
video_formats_tested = global_config.video_formats_tested
# TODO: Check `parse_image` it was written very fast
def play_video(image_source: str | List[ImageSource], add_frame_count:bool=True, parse_image:bool = True) -> None:
    """
    Play `image_source` in an interactive video player implemented entirely within cv2.

    :param image_source: A video path or ImageSource: ndarray, url, path, torch, PIL, video_path. For a full description see: TODO insert link
    :param add_frame_count: If True, add a frame count in the upper left corner.
    :param parse_image: If True, will thoroughly check `image_source` and ensure it's in the correct format
    :return:
    """

    checker.assert_types([add_frame_count, parse_image], [bool, bool], ["add_frame_count", "parse_image"])
    if isinstance(image_source, str): # TODO write a check video string function
        checker.assert_valid_video_path(image_source)
    else:
        if parse_image:
            old_value = global_config.show_max_image_amount
            global_config.show_max_image_amount = 1e6
            image_source = parsers.parse_arbitrary_image_source(image_source)
            global_config.show_max_image_amount = old_value
        if len(image_source) >= 2:
            image_source = image_modifier.center_pad_images_into_nparray(image_source)
    if len(image_source) < 2:
        raise ValueError(f"Expected `source` to contain at least 2 images, but found `{len(image_source)=}`")

    video_player.VideoPlayer(image_source, add_frame_count)


def _extract_frames(cap:cv2.VideoCapture, iterator:ndarray, info:dict, add_frame_count:bool, verbose:bool, frame_skipping_threshold:int) -> ndarray:
    assert len(iterator) > 1, f"The code ly on the fact that there's at least two frame indexes in `{iterator=}`"

    # Initialize return array (I think `np.empty` is the fastest way to do it,  but could be wrong)
    frame_count_goal = len(iterator)
    frames = np.empty((frame_count_goal, info["video_height"], info["video_width"], 3), dtype=np.uint8)
    if verbose:
        iterator = tqdm(iterator)

    try:
        # Extract other frames
        for i, frame_index in enumerate(iterator):
            if i == 0:
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
            elif (frame_index-last_index) < frame_skipping_threshold:
                for _ in range(frame_index - last_index - 1): # read `frame_index-last_index` frames without doing anything to them
                    if not cap.read()[0]:
                        raise RuntimeError("Failed to read data while attempting to skip frames")
            else:
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)

            was_successful, frame = cap.read()
            if not was_successful:
                raise RuntimeError("Unable to read the frame")
            if add_frame_count:
                frame = image_modifier.draw_text_cv2(frame, str(frame_index + 1), position="upper_left_corner", deep_copy=False)
            frames[i, ...] = frame
            last_index = frame_index
    except Exception as e:
        raise RuntimeError(f"Failed reading the `{frame_index}`th frame due to error: `{e}`")

    cap.release()
    assert len(frames.shape) == 4, "Something went wrong"
    return frames[:, :, :, ::-1]  # ::-1 is a fast BGR --> RGB


def parse_video_to_images(video_path: str, keep_nth_frame: int=1, add_frame_count: bool = True, verbose: bool = True, max_frame_count_allowed:int = 500) -> Tuple[np.ndarray, dict]:
    """
    Iterate through `video_path` and extract every `keep_nth_frame` frame.

    :param keep_nth_frame: Number of frames to extract from the video.
    :param add_frame_count: If True, add a frame count in the upper left corner.
    :param verbose: If True, display progress information during video processing.
    :param video_path: Path to the video file.
    :param max_frame_count_allowed: If the combination of `video_path` and `keep_nth_frame` exceeds this limit,
                                    `keep_nth_frame` will be chosen so the amount of frames returned is `max_frame_count_allowed`.
    :returns: (1) a batch of images as a np.ndarray with shape (num_frames, H, W, 3)
              (2) a dict with video info
    """
    # Checks
    checker.assert_types(
        [video_path, keep_nth_frame, add_frame_count, verbose],
        [str, int, bool, bool],
        ["video_path", "keep_nth_frame", "add_frame_count", "verbose"]
    )
    checker.assert_valid_video_path(video_path)
    checker.assert_positive_int(keep_nth_frame, variable_name="keep_nth_frame")

    # Load video
    cap, info = utils.get_cv2_video_capture(video_path)
    iterator = np.arange(0, info["frame_count"], keep_nth_frame)
    frame_count_goal = len(iterator)

    # Frame extraction
    if len(iterator) > max_frame_count_allowed:
        keep_nth_frame = np.ceil(info["frame_count"]/max_frame_count_allowed)
        iterator = np.arange(0, info["frame_count"], keep_nth_frame, dtype="int32")
        msg = f"`{frame_count_goal=}` is higher than the allowed `{max_frame_count_allowed=}`. `keep_nth_frame` will be adjusted to `{keep_nth_frame}`."
        warnings.warn(msg)

    # Frame
    elif (frame_count_goal > 2_000) and (frame_count_goal > max_frame_count_allowed):
        msg = f"Suspiciously large `{frame_count_goal=}` detected. A total of `{frame_count_goal}` " \
              f"frames of shape `{(info['video_height'], info['video_width'])}` will be stored in RAM."
        warnings.warn(msg)

    # Extract frames
    if info["cv2_is_unstable"] and (max(iterator) > 10_000) and (np.max(np.abs(np.diff(iterator))) > 100):
        msg = f"cv2's frame count was unreliable, this prevents speedups during frame iteration."
        warnings.warn(msg)
    t = int(1e6) if info["cv2_is_unstable"] else 100
    frames = _extract_frames(cap, iterator, info, add_frame_count, verbose, frame_skipping_threshold=t)
    return frames, info


def parse_video_to_images_fixed_count(video_path: str, num_frames: int, add_frame_count: bool = True, verbose: bool = True) -> Tuple[np.ndarray, dict]:
    """
    Parse `video_path` into `num_frames` of equally spaced frames.

    :param video_path: Path to the video file.
    :param num_frames: Number of frames to extract from the video.
    :param add_frame_count: If True, add a frame count in the upper left corner.
    :param verbose: If True, display progress information during video processing.
    :returns: (1) a batch of images as a np.ndarray with shape (num_frames, H, W, 3)
              (2) a dict with video info
    """

    # Checks
    checker.assert_types(
        [video_path, num_frames, add_frame_count, verbose],
        [str, int, bool, bool],
        ["video_path", "keep_nth_frame", "add_frame_count", "verbose"]
    )
    checker.assert_valid_video_path(video_path)
    checker.assert_positive_int(num_frames, variable_name="num_frames")

    # Initialize video feed
    cap, info = utils.get_cv2_video_capture(video_path)

    # Determine which frame indexes should be extracted
    all_frame_indexes = np.arange(info["frame_count"])
    iterator = utils.extract_equally_spaced_numbers(all_frame_indexes, num_frames)

    # Extract frames
    # if info["cv2_is_unstable"] and (max(iterator) > 10_000) and (np.max(np.abs(np.diff(iterator))) > 100):
    #     msg = f"cv2's frame count was unreliable, this prevents speedups during frame iteration."
    #     warnings.warn(msg)
    t = int(1e6) if info["cv2_is_unstable"] else 100
    frames = _extract_frames(cap, iterator, info, add_frame_count, verbose, frame_skipping_threshold=t)
    return frames, info


def show_video(
        video_path:str,
        num_frames:int=6,
        add_frame_count:bool=True,
        verbose:bool=True,
        add_video_details:bool=True,
        return_video_details:bool=False,
        title: Optional[str] = None,
        display_image: bool = True,
        return_image: bool = False,
        max_output_image_size_wh: Optional[tuple] = (1920, 1080),
        save_image_path: Optional[str] = None,
        resize_factor: float = 1.0,
        BGR2RGB: bool = False,
) -> Optional[ndarray | Tuple[np.ndarray, dict] | dict]:
    """
    Generate a grid image that displays `num_frames` equally spaced frames from `video_path`.

    :param video_path: Path to the video file.
    :param num_frames: Number of frames that will be displayed (they will be equally spaced across the video, but always include the first and last frame) .
    :param add_frame_count: If True, add frame count to the displayed frames. Default is True.
    :param verbose: If True, display information about the video. Default is True.
    :param add_video_details: If True, add video details (fps, frame size, etc.) to the displayed frames.
    :param return_video_details: If True, return a dict with video details (fps, frame size, etc.)
    :param title: Custom title for the displayed frames. If not provided, the video file name will be used as the title.
    :param display_image: If True, display the processed image using cv2.imshow() or PIL if in jupyter.
    :param return_image: If True, return the processed image as a np.ndarray (H, W, 3).
    :param max_output_image_size_wh: Maximum width and height of the displayed/returned image. If None, nothing will be done.
    :param save_image_path: If provided, save the processed image at the given file path. Default is None (no saving).
    :param resize_factor: Resize factor for each image/frame. Default is 1.0 (no resizing).
    :param BGR2RGB: If True, convert the processed image from BGR to RGB before display/return.

    :returns: If return_image is True, return an image of shape np.ndarray (H, W, 3), otherwise None
    """

    # Checks
    checker.assert_types(
        [video_path, num_frames, add_frame_count, verbose, add_video_details],
        [str, int, bool, bool, bool],
        ["video_path", "keep_nth_frame", "add_frame_count", "verbose", "add_path_to_title"]
    )
    checker.assert_valid_video_path(video_path)
    checker.assert_positive_int(num_frames, variable_name="num_frames")
    checker.assert_valid_universal_inputs(title, display_image, return_image, max_output_image_size_wh, save_image_path, resize_factor, BGR2RGB, True)

    # Parse images and plot them in a grid
    images, info = parse_video_to_images_fixed_count(video_path, num_frames, add_frame_count, verbose)
    final = grid.show_grid(
        image_source=[parsers.resize_and_bgr_parser(image, resize_factor, BGR2RGB) for image in images],
        allow_auto_resize=False,
        title=None,
        display_image=False,
        return_image=True,
        max_output_image_size_wh=max_output_image_size_wh,
        save_image_path=None,
        resize_factor=1.0,
        BGR2RGB=False,
        parse_image=False
    )

    # Title
    if add_video_details:
        video_path_abs = os.path.abspath(video_path)
        final = image_modifier.draw_image_title(final, info["as_string"], top_padding_color=(245, 245, 245), font_size=10, text_color=(100, 100, 100), top_padding=16)
        final = image_modifier.draw_image_title(final, video_path_abs, top_padding_color=(245, 245, 245), font_size=10, text_color=(100, 100, 100), top_padding=16, italic=True)
    if title is not None:
        final = image_modifier.draw_image_title(final, title, top_padding_color=(245, 245, 245))
    if save_image_path is not None:
        utils.save_image_to_disk(save_image_path, final)
    if display_image:
        show.show_image(final)
    if return_image and return_video_details:
        return final, info
    elif return_image and (not return_video_details):
        return final
    elif (not return_image) and return_video_details:
        return info

__all__ = ["play_video", "parse_video_to_images", "parse_video_to_images_fixed_count", "show_video"]

if __name__ == '__main__':
    show_video("C:/Users/Jakob/Desktop/boviwalk_preprocessing/archery.mp4")
    from PIL import Image, ImageDraw
    img = Image.open("hopper.jpg")
    draw = ImageDraw.Draw(img)
    draw.line()


from __future__ import annotations
from typing import Union, List, Tuple, Optional, Iterable, Dict
import numpy as np
import PIL
from PIL import Image
import warnings
import cv2

# A shared, global config file that will only be initialized ones, but that can be edited by the user.
# TODO: I should probably add a shit ton of setters that will warn the user if the input is invalid e.g. `show_max_image_amount<0`
class Config:
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Config, cls).__new__(cls)
        return cls.instance

    # --------------------------------------------
    # Image show and parse
    # --------------------------------------------

    # __show
    show_max_image_amount:int = 100
    image_min_width:int = 5 # This value may not go below 4, because of errors thrown in __parser
    image_min_height:int = 5  # This value may not go below 4, because of errors thrown in __parser

    # Torch
    try_inverse_torch_standardization:bool = True
    torch_standardization_mean:Tuple[int, int, int] = (0.485, 0.456, 0.406)
    torch_standardization_std:Tuple[int, int, int] = (0.229, 0.224, 0.225)

    # numpy
    image_normalization_clip_min:float = -0.01
    image_normalization_clip_max:float = 1.01
    image_clip_min:int = -1
    image_clip_max:int = 1

    # Video
    default_fixed_frame_count_video:int = 6

    # Grid optimization parameters for `show_grid`
    config_auto_grid_layout = {
        "desired_ratio":9/17, # The aspect ratio (h,w) the algorithm aims for
        "ratio_weight":1.0, # How highly difference between the actual and the desired aspect ratio is penalized
        "scale_weight":1.0, # How highly image resizing is penalized
        "empty_weight":1.0  # How highly empty grid cells are penalized
    }

    # --------------------------------------------
    # control
    # --------------------------------------------

    in_debug_mode:bool = False
    cv2_wait_key_delay_ms:int = 0 # 0.0 means it will wait indefinitely, any value greater than zero -> delay in millisecond

    # --------------------------------------------
    # Jupyter
    # --------------------------------------------

    # Check if python is currently running in a jupyter environment. NOTE: This code is somewhat dubious, but I have tested it to the best of my abilities.
    # IN_JUPYTER = False
    # try:
    #     shell = get_ipython().__class__.__name__  # This is supposed to be an unresolved reference anywhere outside jupyter
    #     if shell == 'ZMQInteractiveShell':
    #         IN_JUPYTER = True  # Jupyter notebook or qtconsole
    #     elif shell == 'TerminalInteractiveShell':
    #         IN_JUPYTER = False  # Terminal running IPython (the check is unnecessary, but left it for future reference)
    # except NameError:
    #     # Probably standard Python interpreter
    #     pass

    IN_JUPYTER = False  
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            IN_JUPYTER = True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            IN_JUPYTER = False  # Terminal running IPython (the check is unnecessary, but left it for future reference)
        elif ('IPythonKernel' in shell) or ('ipykernel' in shell):
            IN_JUPYTER = True  # This might catch VSCode's Jupyter notebooks
    except NameError:
        pass

    # --------------------------------------------
    # Image and video formats
    # --------------------------------------------

    video_formats_all:List[str] = ["avi", "mp4", "mkv", "mov", "wmv", "flv", "mpeg", "3gp", "webm", "asf", "vob", "ts", "m4v", "mpg", "rm", "swf", "m2ts", "ogg", "amv"]
    image_formats:list[str] = ["jpg", "jpeg", "png", "bmp", "tiff", "webp"]
    for image_format in image_formats:
        assert image_format not in video_formats_all, f"Overlap between video and image formats: `{image_format}`. This will results in problems later on."
    del image_format
    video_formats_tested:list[str] = ["mp4", "mov"]
    resize_interpolation:int = cv2.INTER_LINEAR #cv2.INTER_CUBIC

    # --------------------------------------------
    # Shared stuff
    # --------------------------------------------

    # So, yeah, this is admittedly an ugly solution.
    # But I want torch type hints e.g. `image:torch.Tensor` even if torch is not installed
    TORCH_FOUND = True
    try:
        import torch
    except ModuleNotFoundError:
        TORCH_FOUND = False

    if IN_JUPYTER:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
    tqdm = tqdm

    if TORCH_FOUND:
        import torch
    else:
        class Torch:
            class dummy_torch_type_its_impossible_to_be_isinstance_off:
                pass
            Tensor = type(dummy_torch_type_its_impossible_to_be_isinstance_off)
        torch = Torch()

    if TORCH_FOUND:
        ImageType = Union[np.ndarray, torch.Tensor, PIL.Image.Image, str]
    else:
        ImageType = Union[np.ndarray, PIL.Image.Image, str]
    ImageSource = Union[np.ndarray, torch.Tensor, ImageType, List[ImageType]]

# Shared stuff
global_config:Config = Config()
ImageType = global_config.ImageType
ImageSource = global_config.ImageSource
ndarray = np.ndarray
torch = global_config.torch
tqdm = global_config.tqdm
TORCH_FOUND:bool = global_config.TORCH_FOUND
IN_JUPYTER:bool = global_config.IN_JUPYTER

__all__ = [
    "global_config", "Dict", "Union", "List", "Tuple", "Optional", "Iterable", "ImageType", "ImageSource", "ndarray", "tqdm", "warnings", "TORCH_FOUND", "IN_JUPYTER", "torch"
]


from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="justshowit",
    url="https://github.com/Jako-K/justshowit",
    version='0.3.6',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Jako-K",
    description='Display images without all the nonsense!',
    packages=find_packages(),
    install_requires=["opencv-python", "requests", "numpy", "Pillow", "tqdm"],
)




